"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 01/04/2017

This module implements a printer for the parsed AST
"""

from lustre2py.visitor import Visitor
from lustre2py.instruction import Equation, Property
from lustre2py.node import Node

class AstPrinter(Visitor):
    """
    Visitor for printing AST on stdout
    """
    def __init__(self):
        pass

    def _visit_node(self, node):
        assert isinstance(node, Node)
        print '\nbegin node', node.name
        if len(node.input_decls) > 0:
            print '\nbegin inputs'
            for inp in node.input_decls:
                inp.accept(self)
            print 'end inputs'
        if len(node.output_decls) > 0:
            print '\nbegin outputs'
            for out in node.output_decls:
                out.accept(self)
            print 'end outputs'
        if len(node.local_decls) > 0:
            print '\nbegin locals'
            for loc in node.local_decls:
                loc.accept(self)
            print 'end locals'
        print '\nbegin equations'
        for equat in node.equations:
            print equat.accept(self)
        print 'end equations'
        if node.main:
            print '--%MAIN;'
        for prop in node.properties:
            print '--%PROPERTY', prop.accept(self) + ';'
        print '\nend node', node.name

    def _visit_var_decl_list(self, lst):
        for grp in lst.var_decl_groups:
            grp.accept(self)

    def _visit_var_decl_group(self, grp):
        var_list = '\t'
        sep = ''
        for var in grp.variables:
            var_list += sep + var
            sep = ', '
        print var_list, ':', grp.datatype.accept(self) + ';'

    def _visit_equation(self, equat):
        assert isinstance(equat, Equation)
        result = '\t'
        sep = ''
        put_par = len(equat.lhs) > 1
        if put_par:
            result += '('
        for lhs in equat.lhs:
            result += sep + lhs
            sep = ', '
        if put_par:
            result += ')'
        result += ' = '
        result += equat.expression.accept(self) + ';'
        return result

    def _visit_init_curr_expression(self, expr):
        result = expr.init.accept(self)
        result += ' -> '
        result += expr.curr.accept(self)
        return result

    def _visit_literal_expression(self, expr):
        return expr.value

    def _visit_zeroary_expression(self, expr):
        return expr.name

    def _visit_unary_expression(self, expr):
        result = expr.operator + ' '
        result += expr.operand.accept(self)
        return result

    def _visit_binary_expression(self, expr):
        result = expr.left.accept(self)
        result += ' ' + expr.operator + ' '
        result += expr.right.accept(self)
        return result

    def _visit_ite_expression(self, expr):
        result = 'if '
        result += expr.if_.accept(self)
        result += ' then '
        result += expr.then_.accept(self)
        result += ' else '
        result += expr.else_.accept(self)
        return result

    def _visit_call_expression(self, expr):
        result = expr.cid
        result += '('
        sep = ''
        for param in expr.params:
            result += sep + param.accept(self)
            sep = ', '
        result += ')'
        return result

    def _visit_primitive(self, obj):
        return obj.name

    def _visit_tuple_expression(self, ttuple):
        result = '('
        sep = ''
        for expr in ttuple.expressions:
            result += sep + expr.accept(self)
            sep = ', '
        result += ')'
        return result

    def _visit_enum(self, obj):
        raise NotImplementedError

    def _visit_struct(self, obj):
        raise NotImplementedError

    def _visit_array(self, obj):
        raise NotImplementedError

    def _visit_subrange(self, obj):
        raise NotImplementedError

    def _visit_property(self, obj):
        assert isinstance(obj, Property)
        return obj.expression.accept(self)



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017
"""
import intrepyd as ip
import intrepyd.scr
import intrepyd.circuit
import collections

class SimulinkCircuit(ip.circuit.Circuit):
    def __init__(self, ctx, name):
        ip.circuit.Circuit.__init__(self, ctx, name)

    def _mk_naked_circuit_impl(self, inputs):
        input_keys = list(inputs)
        # OnOff -> n1
        n1 = inputs[input_keys[0]]
        # OnGreen -> n2
        n2 = inputs[input_keys[1]]
        # OnYellow -> n3
        n3 = inputs[input_keys[2]]
        # OnRed -> n4
        n4 = inputs[input_keys[3]]
        # Daytime -> n5
        n5 = inputs[input_keys[4]]
        n6 = self.context.mk_latch('traffic_light/Past(OnOff)', self.context.mk_boolean_type())
        n7 = self.context.mk_latch('traffic_light/Past(OnRed)', self.context.mk_boolean_type())
        n8 = self.context.mk_latch('traffic_light/Past(OnYellow)', self.context.mk_boolean_type())
        n9 = self.context.mk_latch('traffic_light/Past(OnGreen)', self.context.mk_boolean_type())
        n10 = self.context.mk_latch('traffic_light/Past(Daytime)', self.context.mk_boolean_type())
        n11 = self.context.mk_latch('traffic_light/Init', self.context.mk_boolean_type())
        n12 = self.context.mk_latch('traffic_light/Past(Mode)', self.context.mk_int8_type())
        # traffic_light/false
        n13 = self.context.mk_false()
        self.nets['traffic_light/false'] = n13
        # traffic_light/Off
        n14 = self.context.mk_number('0', self.context.mk_int8_type())
        self.nets['traffic_light/Off'] = n14
        # Bus Creator
        n15 = [n1, n2, n3, n4, n5]
        # Bus Creator1
        n16 = [n6, n9, n8, n7, n10]
        # traffic_light/Green
        n17 = self.context.mk_number('1', self.context.mk_int8_type())
        self.nets['traffic_light/Green'] = n17
        # traffic_light/Yellow
        n18 = self.context.mk_number('2', self.context.mk_int8_type())
        self.nets['traffic_light/Yellow'] = n18
        # traffic_light/Red
        n19 = self.context.mk_number('3', self.context.mk_int8_type())
        self.nets['traffic_light/Red'] = n19
        # Bus Creator2
        n20 = [n14, n17, n18, n19]
        # traffic_light/Mode1
        n21 = self.context.mk_ite(n11, n14, n12)
        self.nets['traffic_light/Mode1'] = n21
        n22_1 = ip.scr.mk_scr(self.context, 'tests/traffic_light', n15, n16, n20, n21)
        # traffic_light/Mode
        n23 = self.context.mk_ite(n11, n14, n22_1)
        self.nets['traffic_light/Mode'] = n23
        in6 = self.context.mk_true()
        self.context.set_latch_init_next(n6, in6, n1)
        in7 = self.context.mk_true()
        self.context.set_latch_init_next(n7, in7, n4)
        in8 = self.context.mk_true()
        self.context.set_latch_init_next(n8, in8, n3)
        in9 = self.context.mk_true()
        self.context.set_latch_init_next(n9, in9, n2)
        in10 = self.context.mk_true()
        self.context.set_latch_init_next(n10, in10, n5)
        in11 = self.context.mk_true()
        self.context.set_latch_init_next(n11, in11, n13)
        in12 = self.context.mk_number('1', self.context.mk_int8_type())
        self.context.set_latch_init_next(n12, in12, n23)
        # n23 -> out
        outputs = collections.OrderedDict()
        outputs['traffic_light/out'] = n23
        self.nets['traffic_light/out'] = n23
        return outputs

    def _mk_inputs(self):
        # traffic_light/OnOff -> n1
        n1 = self.context.mk_input('OnOff', self.context.mk_boolean_type())
        self.inputs['OnOff'] = n1
        # traffic_light/OnGreen -> n2
        n2 = self.context.mk_input('OnGreen', self.context.mk_boolean_type())
        self.inputs['OnGreen'] = n2
        # traffic_light/OnYellow -> n3
        n3 = self.context.mk_input('OnYellow', self.context.mk_boolean_type())
        self.inputs['OnYellow'] = n3
        # traffic_light/OnRed -> n4
        n4 = self.context.mk_input('OnRed', self.context.mk_boolean_type())
        self.inputs['OnRed'] = n4
        # traffic_light/Daytime -> n5
        n5 = self.context.mk_input('Daytime', self.context.mk_boolean_type())
        self.inputs['Daytime'] = n5




# Generated from Lustre.g4 by ANTLR 4.6
# encoding: utf-8
from __future__ import print_function
from antlr4 import *
from io import StringIO

def serializedATN():
    with StringIO() as buf:
        buf.write(u"\3\u0430\ud6d1\u8206\uad2d\u4417\uaef1\u8d80\uaadd\3")
        buf.write(u"A\u0191\4\2\t\2\4\3\t\3\4\4\t\4\4\5\t\5\4\6\t\6\4\7\t")
        buf.write(u"\7\4\b\t\b\4\t\t\t\4\n\t\n\4\13\t\13\4\f\t\f\4\r\t\r")
        buf.write(u"\4\16\t\16\4\17\t\17\4\20\t\20\4\21\t\21\4\22\t\22\4")
        buf.write(u"\23\t\23\4\24\t\24\4\25\t\25\3\2\3\2\3\2\7\2.\n\2\f\2")
        buf.write(u"\16\2\61\13\2\3\2\3\2\3\3\3\3\3\3\3\3\3\3\3\3\3\4\3\4")
        buf.write(u"\3\4\3\4\5\4?\n\4\3\4\3\4\3\4\3\4\3\5\3\5\3\5\3\5\5\5")
        buf.write(u"I\n\5\3\5\3\5\3\5\3\5\5\5O\n\5\3\5\3\5\3\5\3\5\3\5\3")
        buf.write(u"\5\5\5W\n\5\3\5\3\5\5\5[\n\5\3\5\3\5\5\5_\n\5\3\6\6\6")
        buf.write(u"b\n\6\r\6\16\6c\3\7\3\7\3\7\3\7\3\7\3\7\5\7l\n\7\3\b")
        buf.write(u"\3\b\3\b\7\bq\n\b\f\b\16\bt\13\b\3\t\3\t\3\t\7\ty\n\t")
        buf.write(u"\f\t\16\t|\13\t\3\t\3\t\3\t\3\n\3\n\3\n\3\n\3\n\3\n\3")
        buf.write(u"\n\3\n\3\n\3\n\3\n\7\n\u008c\n\n\f\n\16\n\u008f\13\n")
        buf.write(u"\3\n\3\n\3\n\3\n\3\n\3\n\3\n\7\n\u0098\n\n\f\n\16\n\u009b")
        buf.write(u"\13\n\3\n\5\n\u009e\n\n\3\13\3\13\3\13\3\13\3\13\3\13")
        buf.write(u"\3\13\3\13\3\13\3\13\3\13\3\13\3\13\3\13\5\13\u00ae\n")
        buf.write(u"\13\3\13\3\13\3\13\3\13\7\13\u00b4\n\13\f\13\16\13\u00b7")
        buf.write(u"\13\13\3\f\5\f\u00ba\n\f\3\f\3\f\3\r\3\r\3\r\3\r\3\16")
        buf.write(u"\3\16\3\16\3\16\7\16\u00c6\n\16\f\16\16\16\u00c9\13\16")
        buf.write(u"\5\16\u00cb\n\16\3\16\3\16\3\17\3\17\3\17\3\17\7\17\u00d3")
        buf.write(u"\n\17\f\17\16\17\u00d6\13\17\5\17\u00d8\n\17\3\17\3\17")
        buf.write(u"\3\20\3\20\5\20\u00de\n\20\3\21\3\21\3\21\3\21\3\22\3")
        buf.write(u"\22\3\22\5\22\u00e7\n\22\3\22\5\22\u00ea\n\22\3\22\3")
        buf.write(u"\22\3\22\3\22\3\23\3\23\3\23\7\23\u00f3\n\23\f\23\16")
        buf.write(u"\23\u00f6\13\23\3\24\3\24\3\24\5\24\u00fb\n\24\3\24\3")
        buf.write(u"\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24")
        buf.write(u"\3\24\7\24\u010a\n\24\f\24\16\24\u010d\13\24\5\24\u010f")
        buf.write(u"\n\24\3\24\3\24\3\24\3\24\3\24\3\24\6\24\u0117\n\24\r")
        buf.write(u"\24\16\24\u0118\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3")
        buf.write(u"\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24")
        buf.write(u"\3\24\3\24\3\24\3\24\3\24\3\24\7\24\u0133\n\24\f\24\16")
        buf.write(u"\24\u0136\13\24\3\24\3\24\3\24\3\24\3\24\3\24\7\24\u013e")
        buf.write(u"\n\24\f\24\16\24\u0141\13\24\3\24\3\24\3\24\3\24\3\24")
        buf.write(u"\3\24\7\24\u0149\n\24\f\24\16\24\u014c\13\24\3\24\3\24")
        buf.write(u"\5\24\u0150\n\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3")
        buf.write(u"\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24")
        buf.write(u"\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3")
        buf.write(u"\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24\3\24")
        buf.write(u"\3\24\3\24\3\24\3\24\7\24\u017d\n\24\f\24\16\24\u0180")
        buf.write(u"\13\24\3\25\3\25\3\25\3\25\3\25\3\25\3\25\3\25\3\25\3")
        buf.write(u"\25\7\25\u018c\n\25\f\25\16\25\u018f\13\25\3\25\2\5\24")
        buf.write(u"&(\26\2\4\6\b\n\f\16\20\22\24\26\30\32\34\36 \"$&(\2")
        buf.write(u"\7\4\2\32\32!!\3\2\'*\4\2\33\33++\4\2\4\4,\60\3\2\62")
        buf.write(u"\63\u01c0\2/\3\2\2\2\4\64\3\2\2\2\6:\3\2\2\2\bD\3\2\2")
        buf.write(u"\2\na\3\2\2\2\fk\3\2\2\2\16m\3\2\2\2\20u\3\2\2\2\22\u009d")
        buf.write(u"\3\2\2\2\24\u00ad\3\2\2\2\26\u00b9\3\2\2\2\30\u00bd\3")
        buf.write(u"\2\2\2\32\u00c1\3\2\2\2\34\u00ce\3\2\2\2\36\u00db\3\2")
        buf.write(u"\2\2 \u00df\3\2\2\2\"\u00e9\3\2\2\2$\u00ef\3\2\2\2&\u014f")
        buf.write(u"\3\2\2\2(\u0181\3\2\2\2*.\5\4\3\2+.\5\6\4\2,.\5\b\5\2")
        buf.write(u"-*\3\2\2\2-+\3\2\2\2-,\3\2\2\2.\61\3\2\2\2/-\3\2\2\2")
        buf.write(u"/\60\3\2\2\2\60\62\3\2\2\2\61/\3\2\2\2\62\63\7\2\2\3")
        buf.write(u"\63\3\3\2\2\2\64\65\7\3\2\2\65\66\7<\2\2\66\67\7\4\2")
        buf.write(u"\2\678\5\22\n\289\7\5\2\29\5\3\2\2\2:;\7\6\2\2;>\7<\2")
        buf.write(u"\2<=\7\7\2\2=?\5\24\13\2><\3\2\2\2>?\3\2\2\2?@\3\2\2")
        buf.write(u"\2@A\7\4\2\2AB\5&\24\2BC\7\5\2\2C\7\3\2\2\2DE\7\b\2\2")
        buf.write(u"EF\7<\2\2FH\7\t\2\2GI\5\16\b\2HG\3\2\2\2HI\3\2\2\2IJ")
        buf.write(u"\3\2\2\2JK\7\n\2\2KL\7\13\2\2LN\7\t\2\2MO\5\16\b\2NM")
        buf.write(u"\3\2\2\2NO\3\2\2\2OP\3\2\2\2PQ\7\n\2\2QV\7\5\2\2RS\7")
        buf.write(u"\f\2\2ST\5\16\b\2TU\7\5\2\2UW\3\2\2\2VR\3\2\2\2VW\3\2")
        buf.write(u"\2\2WX\3\2\2\2XZ\7\r\2\2Y[\5\n\6\2ZY\3\2\2\2Z[\3\2\2")
        buf.write(u"\2[\\\3\2\2\2\\^\7\16\2\2]_\7\5\2\2^]\3\2\2\2^_\3\2\2")
        buf.write(u"\2_\t\3\2\2\2`b\5\f\7\2a`\3\2\2\2bc\3\2\2\2ca\3\2\2\2")
        buf.write(u"cd\3\2\2\2d\13\3\2\2\2el\5\"\22\2fl\5\30\r\2gl\5 \21")
        buf.write(u"\2hl\5\36\20\2il\5\32\16\2jl\5\34\17\2ke\3\2\2\2kf\3")
        buf.write(u"\2\2\2kg\3\2\2\2kh\3\2\2\2ki\3\2\2\2kj\3\2\2\2l\r\3\2")
        buf.write(u"\2\2mr\5\20\t\2no\7\5\2\2oq\5\20\t\2pn\3\2\2\2qt\3\2")
        buf.write(u"\2\2rp\3\2\2\2rs\3\2\2\2s\17\3\2\2\2tr\3\2\2\2uz\5(\25")
        buf.write(u"\2vw\7\17\2\2wy\5(\25\2xv\3\2\2\2y|\3\2\2\2zx\3\2\2\2")
        buf.write(u"z{\3\2\2\2{}\3\2\2\2|z\3\2\2\2}~\7\7\2\2~\177\5\24\13")
        buf.write(u"\2\177\21\3\2\2\2\u0080\u009e\5\24\13\2\u0081\u0082\7")
        buf.write(u"\20\2\2\u0082\u0083\7\21\2\2\u0083\u0084\7<\2\2\u0084")
        buf.write(u"\u0085\7\7\2\2\u0085\u0086\5\24\13\2\u0086\u008d\3\2")
        buf.write(u"\2\2\u0087\u0088\7\5\2\2\u0088\u0089\7<\2\2\u0089\u008a")
        buf.write(u"\7\7\2\2\u008a\u008c\5\24\13\2\u008b\u0087\3\2\2\2\u008c")
        buf.write(u"\u008f\3\2\2\2\u008d\u008b\3\2\2\2\u008d\u008e\3\2\2")
        buf.write(u"\2\u008e\u0090\3\2\2\2\u008f\u008d\3\2\2\2\u0090\u0091")
        buf.write(u"\7\22\2\2\u0091\u009e\3\2\2\2\u0092\u0093\7\23\2\2\u0093")
        buf.write(u"\u0094\7\21\2\2\u0094\u0099\7<\2\2\u0095\u0096\7\17\2")
        buf.write(u"\2\u0096\u0098\7<\2\2\u0097\u0095\3\2\2\2\u0098\u009b")
        buf.write(u"\3\2\2\2\u0099\u0097\3\2\2\2\u0099\u009a\3\2\2\2\u009a")
        buf.write(u"\u009c\3\2\2\2\u009b\u0099\3\2\2\2\u009c\u009e\7\22\2")
        buf.write(u"\2\u009d\u0080\3\2\2\2\u009d\u0081\3\2\2\2\u009d\u0092")
        buf.write(u"\3\2\2\2\u009e\23\3\2\2\2\u009f\u00a0\b\13\1\2\u00a0")
        buf.write(u"\u00ae\7\24\2\2\u00a1\u00a2\7\25\2\2\u00a2\u00a3\7\26")
        buf.write(u"\2\2\u00a3\u00a4\5\26\f\2\u00a4\u00a5\7\17\2\2\u00a5")
        buf.write(u"\u00a6\5\26\f\2\u00a6\u00a7\7\27\2\2\u00a7\u00a8\7\30")
        buf.write(u"\2\2\u00a8\u00a9\7\24\2\2\u00a9\u00ae\3\2\2\2\u00aa\u00ae")
        buf.write(u"\7\31\2\2\u00ab\u00ae\7\32\2\2\u00ac\u00ae\7<\2\2\u00ad")
        buf.write(u"\u009f\3\2\2\2\u00ad\u00a1\3\2\2\2\u00ad\u00aa\3\2\2")
        buf.write(u"\2\u00ad\u00ab\3\2\2\2\u00ad\u00ac\3\2\2\2\u00ae\u00b5")
        buf.write(u"\3\2\2\2\u00af\u00b0\f\4\2\2\u00b0\u00b1\7\26\2\2\u00b1")
        buf.write(u"\u00b2\7;\2\2\u00b2\u00b4\7\27\2\2\u00b3\u00af\3\2\2")
        buf.write(u"\2\u00b4\u00b7\3\2\2\2\u00b5\u00b3\3\2\2\2\u00b5\u00b6")
        buf.write(u"\3\2\2\2\u00b6\25\3\2\2\2\u00b7\u00b5\3\2\2\2\u00b8\u00ba")
        buf.write(u"\7\33\2\2\u00b9\u00b8\3\2\2\2\u00b9\u00ba\3\2\2\2\u00ba")
        buf.write(u"\u00bb\3\2\2\2\u00bb\u00bc\7;\2\2\u00bc\27\3\2\2\2\u00bd")
        buf.write(u"\u00be\7\34\2\2\u00be\u00bf\5&\24\2\u00bf\u00c0\7\5\2")
        buf.write(u"\2\u00c0\31\3\2\2\2\u00c1\u00ca\7\35\2\2\u00c2\u00c7")
        buf.write(u"\7<\2\2\u00c3\u00c4\7\17\2\2\u00c4\u00c6\7<\2\2\u00c5")
        buf.write(u"\u00c3\3\2\2\2\u00c6\u00c9\3\2\2\2\u00c7\u00c5\3\2\2")
        buf.write(u"\2\u00c7\u00c8\3\2\2\2\u00c8\u00cb\3\2\2\2\u00c9\u00c7")
        buf.write(u"\3\2\2\2\u00ca\u00c2\3\2\2\2\u00ca\u00cb\3\2\2\2\u00cb")
        buf.write(u"\u00cc\3\2\2\2\u00cc\u00cd\7\5\2\2\u00cd\33\3\2\2\2\u00ce")
        buf.write(u"\u00d7\7\36\2\2\u00cf\u00d4\7<\2\2\u00d0\u00d1\7\17\2")
        buf.write(u"\2\u00d1\u00d3\7<\2\2\u00d2\u00d0\3\2\2\2\u00d3\u00d6")
        buf.write(u"\3\2\2\2\u00d4\u00d2\3\2\2\2\u00d4\u00d5\3\2\2\2\u00d5")
        buf.write(u"\u00d8\3\2\2\2\u00d6\u00d4\3\2\2\2\u00d7\u00cf\3\2\2")
        buf.write(u"\2\u00d7\u00d8\3\2\2\2\u00d8\u00d9\3\2\2\2\u00d9\u00da")
        buf.write(u"\7\5\2\2\u00da\35\3\2\2\2\u00db\u00dd\7\37\2\2\u00dc")
        buf.write(u"\u00de\7\5\2\2\u00dd\u00dc\3\2\2\2\u00dd\u00de\3\2\2")
        buf.write(u"\2\u00de\37\3\2\2\2\u00df\u00e0\7 \2\2\u00e0\u00e1\5")
        buf.write(u"&\24\2\u00e1\u00e2\7\5\2\2\u00e2!\3\2\2\2\u00e3\u00ea")
        buf.write(u"\5$\23\2\u00e4\u00e6\7\t\2\2\u00e5\u00e7\5$\23\2\u00e6")
        buf.write(u"\u00e5\3\2\2\2\u00e6\u00e7\3\2\2\2\u00e7\u00e8\3\2\2")
        buf.write(u"\2\u00e8\u00ea\7\n\2\2\u00e9\u00e3\3\2\2\2\u00e9\u00e4")
        buf.write(u"\3\2\2\2\u00ea\u00eb\3\2\2\2\u00eb\u00ec\7\4\2\2\u00ec")
        buf.write(u"\u00ed\5&\24\2\u00ed\u00ee\7\5\2\2\u00ee#\3\2\2\2\u00ef")
        buf.write(u"\u00f4\5(\25\2\u00f0\u00f1\7\17\2\2\u00f1\u00f3\5(\25")
        buf.write(u"\2\u00f2\u00f0\3\2\2\2\u00f3\u00f6\3\2\2\2\u00f4\u00f2")
        buf.write(u"\3\2\2\2\u00f4\u00f5\3\2\2\2\u00f5%\3\2\2\2\u00f6\u00f4")
        buf.write(u"\3\2\2\2\u00f7\u00f8\b\24\1\2\u00f8\u0150\7<\2\2\u00f9")
        buf.write(u"\u00fb\7\33\2\2\u00fa\u00f9\3\2\2\2\u00fa\u00fb\3\2\2")
        buf.write(u"\2\u00fb\u00fc\3\2\2\2\u00fc\u0150\7;\2\2\u00fd\u0150")
        buf.write(u"\79\2\2\u00fe\u0150\7:\2\2\u00ff\u0100\t\2\2\2\u0100")
        buf.write(u"\u0101\7\t\2\2\u0101\u0102\5&\24\2\u0102\u0103\7\n\2")
        buf.write(u"\2\u0103\u0150\3\2\2\2\u0104\u0105\7<\2\2\u0105\u010e")
        buf.write(u"\7\t\2\2\u0106\u010b\5&\24\2\u0107\u0108\7\17\2\2\u0108")
        buf.write(u"\u010a\5&\24\2\u0109\u0107\3\2\2\2\u010a\u010d\3\2\2")
        buf.write(u"\2\u010b\u0109\3\2\2\2\u010b\u010c\3\2\2\2\u010c\u010f")
        buf.write(u"\3\2\2\2\u010d\u010b\3\2\2\2\u010e\u0106\3\2\2\2\u010e")
        buf.write(u"\u010f\3\2\2\2\u010f\u0110\3\2\2\2\u0110\u0150\7\n\2")
        buf.write(u"\2\u0111\u0112\7\"\2\2\u0112\u0113\7\t\2\2\u0113\u0116")
        buf.write(u"\5&\24\2\u0114\u0115\7\17\2\2\u0115\u0117\5&\24\2\u0116")
        buf.write(u"\u0114\3\2\2\2\u0117\u0118\3\2\2\2\u0118\u0116\3\2\2")
        buf.write(u"\2\u0118\u0119\3\2\2\2\u0119\u011a\3\2\2\2\u011a\u011b")
        buf.write(u"\7\n\2\2\u011b\u0150\3\2\2\2\u011c\u011d\7%\2\2\u011d")
        buf.write(u"\u0150\5&\24\20\u011e\u011f\7&\2\2\u011f\u0150\5&\24")
        buf.write(u"\17\u0120\u0121\7\33\2\2\u0121\u0150\5&\24\16\u0122\u0123")
        buf.write(u"\7\66\2\2\u0123\u0124\5&\24\2\u0124\u0125\7\67\2\2\u0125")
        buf.write(u"\u0126\5&\24\2\u0126\u0127\78\2\2\u0127\u0128\5&\24\6")
        buf.write(u"\u0128\u0150\3\2\2\2\u0129\u012a\7<\2\2\u012a\u012b\7")
        buf.write(u"\21\2\2\u012b\u012c\7<\2\2\u012c\u012d\7\4\2\2\u012d")
        buf.write(u"\u0134\5&\24\2\u012e\u012f\7\5\2\2\u012f\u0130\7<\2\2")
        buf.write(u"\u0130\u0131\7\4\2\2\u0131\u0133\5&\24\2\u0132\u012e")
        buf.write(u"\3\2\2\2\u0133\u0136\3\2\2\2\u0134\u0132\3\2\2\2\u0134")
        buf.write(u"\u0135\3\2\2\2\u0135\u0137\3\2\2\2\u0136\u0134\3\2\2")
        buf.write(u"\2\u0137\u0138\7\22\2\2\u0138\u0150\3\2\2\2\u0139\u013a")
        buf.write(u"\7\26\2\2\u013a\u013f\5&\24\2\u013b\u013c\7\17\2\2\u013c")
        buf.write(u"\u013e\5&\24\2\u013d\u013b\3\2\2\2\u013e\u0141\3\2\2")
        buf.write(u"\2\u013f\u013d\3\2\2\2\u013f\u0140\3\2\2\2\u0140\u0142")
        buf.write(u"\3\2\2\2\u0141\u013f\3\2\2\2\u0142\u0143\7\27\2\2\u0143")
        buf.write(u"\u0150\3\2\2\2\u0144\u0145\7\t\2\2\u0145\u014a\5&\24")
        buf.write(u"\2\u0146\u0147\7\17\2\2\u0147\u0149\5&\24\2\u0148\u0146")
        buf.write(u"\3\2\2\2\u0149\u014c\3\2\2\2\u014a\u0148\3\2\2\2\u014a")
        buf.write(u"\u014b\3\2\2\2\u014b\u014d\3\2\2\2\u014c\u014a\3\2\2")
        buf.write(u"\2\u014d\u014e\7\n\2\2\u014e\u0150\3\2\2\2\u014f\u00f7")
        buf.write(u"\3\2\2\2\u014f\u00fa\3\2\2\2\u014f\u00fd\3\2\2\2\u014f")
        buf.write(u"\u00fe\3\2\2\2\u014f\u00ff\3\2\2\2\u014f\u0104\3\2\2")
        buf.write(u"\2\u014f\u0111\3\2\2\2\u014f\u011c\3\2\2\2\u014f\u011e")
        buf.write(u"\3\2\2\2\u014f\u0120\3\2\2\2\u014f\u0122\3\2\2\2\u014f")
        buf.write(u"\u0129\3\2\2\2\u014f\u0139\3\2\2\2\u014f\u0144\3\2\2")
        buf.write(u"\2\u0150\u017e\3\2\2\2\u0151\u0152\f\r\2\2\u0152\u0153")
        buf.write(u"\t\3\2\2\u0153\u017d\5&\24\16\u0154\u0155\f\f\2\2\u0155")
        buf.write(u"\u0156\t\4\2\2\u0156\u017d\5&\24\r\u0157\u0158\f\13\2")
        buf.write(u"\2\u0158\u0159\t\5\2\2\u0159\u017d\5&\24\f\u015a\u015b")
        buf.write(u"\f\n\2\2\u015b\u015c\7\61\2\2\u015c\u017d\5&\24\13\u015d")
        buf.write(u"\u015e\f\t\2\2\u015e\u015f\t\6\2\2\u015f\u017d\5&\24")
        buf.write(u"\n\u0160\u0161\f\b\2\2\u0161\u0162\7\64\2\2\u0162\u017d")
        buf.write(u"\5&\24\b\u0163\u0164\f\7\2\2\u0164\u0165\7\65\2\2\u0165")
        buf.write(u"\u017d\5&\24\7\u0166\u0167\f\24\2\2\u0167\u0168\7#\2")
        buf.write(u"\2\u0168\u017d\7<\2\2\u0169\u016a\f\23\2\2\u016a\u016b")
        buf.write(u"\7\21\2\2\u016b\u016c\7<\2\2\u016c\u016d\7$\2\2\u016d")
        buf.write(u"\u016e\5&\24\2\u016e\u016f\7\22\2\2\u016f\u017d\3\2\2")
        buf.write(u"\2\u0170\u0171\f\22\2\2\u0171\u0172\7\26\2\2\u0172\u0173")
        buf.write(u"\5&\24\2\u0173\u0174\7\27\2\2\u0174\u017d\3\2\2\2\u0175")
        buf.write(u"\u0176\f\21\2\2\u0176\u0177\7\26\2\2\u0177\u0178\5&\24")
        buf.write(u"\2\u0178\u0179\7$\2\2\u0179\u017a\5&\24\2\u017a\u017b")
        buf.write(u"\7\27\2\2\u017b\u017d\3\2\2\2\u017c\u0151\3\2\2\2\u017c")
        buf.write(u"\u0154\3\2\2\2\u017c\u0157\3\2\2\2\u017c\u015a\3\2\2")
        buf.write(u"\2\u017c\u015d\3\2\2\2\u017c\u0160\3\2\2\2\u017c\u0163")
        buf.write(u"\3\2\2\2\u017c\u0166\3\2\2\2\u017c\u0169\3\2\2\2\u017c")
        buf.write(u"\u0170\3\2\2\2\u017c\u0175\3\2\2\2\u017d\u0180\3\2\2")
        buf.write(u"\2\u017e\u017c\3\2\2\2\u017e\u017f\3\2\2\2\u017f\'\3")
        buf.write(u"\2\2\2\u0180\u017e\3\2\2\2\u0181\u0182\b\25\1\2\u0182")
        buf.write(u"\u0183\7<\2\2\u0183\u018d\3\2\2\2\u0184\u0185\f\4\2\2")
        buf.write(u"\u0185\u0186\7\26\2\2\u0186\u0187\7;\2\2\u0187\u018c")
        buf.write(u"\7\27\2\2\u0188\u0189\f\3\2\2\u0189\u018a\7#\2\2\u018a")
        buf.write(u"\u018c\7<\2\2\u018b\u0184\3\2\2\2\u018b\u0188\3\2\2\2")
        buf.write(u"\u018c\u018f\3\2\2\2\u018d\u018b\3\2\2\2\u018d\u018e")
        buf.write(u"\3\2\2\2\u018e)\3\2\2\2\u018f\u018d\3\2\2\2(-/>HNVZ^")
        buf.write(u"ckrz\u008d\u0099\u009d\u00ad\u00b5\u00b9\u00c7\u00ca")
        buf.write(u"\u00d4\u00d7\u00dd\u00e6\u00e9\u00f4\u00fa\u010b\u010e")
        buf.write(u"\u0118\u0134\u013f\u014a\u014f\u017c\u017e\u018b\u018d")
        return buf.getvalue()


class LustreParser ( Parser ):

    grammarFileName = "Lustre.g4"

    atn = ATNDeserializer().deserialize(serializedATN())

    decisionsToDFA = [ DFA(ds, i) for i, ds in enumerate(atn.decisionToState) ]

    sharedContextCache = PredictionContextCache()

    literalNames = [ u"<INVALID>", u"'datatype'", u"'='", u"';'", u"'const'", 
                     u"':'", u"'node'", u"'('", u"')'", u"'returns'", u"'var'", 
                     u"'let'", u"'tel'", u"','", u"'struct'", u"'{'", u"'}'", 
                     u"'enum'", u"'int'", u"'subrange'", u"'['", u"']'", 
                     u"'of'", u"'bool'", u"'real'", u"'-'", u"'--%PROPERTY'", 
                     u"'--%REALIZABLE'", u"'--%IVC'", u"'--%MAIN'", u"'assert'", 
                     u"'floor'", u"'condact'", u"'.'", u"':='", u"'pre'", 
                     u"'not'", u"'*'", u"'/'", u"'div'", u"'mod'", u"'+'", 
                     u"'<'", u"'<='", u"'>'", u"'>='", u"'<>'", u"'and'", 
                     u"'or'", u"'xor'", u"'=>'", u"'->'", u"'if'", u"'then'", 
                     u"'else'" ]

    symbolicNames = [ u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"<INVALID>", 
                      u"<INVALID>", u"<INVALID>", u"<INVALID>", u"REAL", 
                      u"BOOL", u"INT", u"ID", u"WS", u"SL_COMMENT", u"ML_COMMENT", 
                      u"C_COMMENT", u"ERROR" ]

    RULE_program = 0
    RULE_datatypedef = 1
    RULE_constant = 2
    RULE_node = 3
    RULE_instrList = 4
    RULE_instr = 5
    RULE_varDeclList = 6
    RULE_varDeclGroup = 7
    RULE_topLevelType = 8
    RULE_datatype = 9
    RULE_bound = 10
    RULE_prop = 11
    RULE_realizabilityInputs = 12
    RULE_ivc = 13
    RULE_main = 14
    RULE_assertion = 15
    RULE_equation = 16
    RULE_lhs = 17
    RULE_expr = 18
    RULE_eID = 19

    ruleNames =  [ u"program", u"datatypedef", u"constant", u"node", u"instrList", 
                   u"instr", u"varDeclList", u"varDeclGroup", u"topLevelType", 
                   u"datatype", u"bound", u"prop", u"realizabilityInputs", 
                   u"ivc", u"main", u"assertion", u"equation", u"lhs", u"expr", 
                   u"eID" ]

    EOF = Token.EOF
    T__0=1
    T__1=2
    T__2=3
    T__3=4
    T__4=5
    T__5=6
    T__6=7
    T__7=8
    T__8=9
    T__9=10
    T__10=11
    T__11=12
    T__12=13
    T__13=14
    T__14=15
    T__15=16
    T__16=17
    T__17=18
    T__18=19
    T__19=20
    T__20=21
    T__21=22
    T__22=23
    T__23=24
    T__24=25
    T__25=26
    T__26=27
    T__27=28
    T__28=29
    T__29=30
    T__30=31
    T__31=32
    T__32=33
    T__33=34
    T__34=35
    T__35=36
    T__36=37
    T__37=38
    T__38=39
    T__39=40
    T__40=41
    T__41=42
    T__42=43
    T__43=44
    T__44=45
    T__45=46
    T__46=47
    T__47=48
    T__48=49
    T__49=50
    T__50=51
    T__51=52
    T__52=53
    T__53=54
    REAL=55
    BOOL=56
    INT=57
    ID=58
    WS=59
    SL_COMMENT=60
    ML_COMMENT=61
    C_COMMENT=62
    ERROR=63

    def __init__(self, input):
        super(LustreParser, self).__init__(input)
        self.checkVersion("4.6")
        self._interp = ParserATNSimulator(self, self.atn, self.decisionsToDFA, self.sharedContextCache)
        self._predicates = None



    class ProgramContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.ProgramContext, self).__init__(parent, invokingState)
            self.parser = parser

        def EOF(self):
            return self.getToken(LustreParser.EOF, 0)

        def datatypedef(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.DatatypedefContext)
            else:
                return self.getTypedRuleContext(LustreParser.DatatypedefContext,i)


        def constant(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ConstantContext)
            else:
                return self.getTypedRuleContext(LustreParser.ConstantContext,i)


        def node(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.NodeContext)
            else:
                return self.getTypedRuleContext(LustreParser.NodeContext,i)


        def getRuleIndex(self):
            return LustreParser.RULE_program

        def accept(self, visitor):
            if hasattr(visitor, "visitProgram"):
                return visitor.visitProgram(self)
            else:
                return visitor.visitChildren(self)




    def program(self):

        localctx = LustreParser.ProgramContext(self, self._ctx, self.state)
        self.enterRule(localctx, 0, self.RULE_program)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 45
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << LustreParser.T__0) | (1 << LustreParser.T__3) | (1 << LustreParser.T__5))) != 0):
                self.state = 43
                self._errHandler.sync(self)
                token = self._input.LA(1)
                if token in [LustreParser.T__0]:
                    self.state = 40
                    self.datatypedef()
                    pass
                elif token in [LustreParser.T__3]:
                    self.state = 41
                    self.constant()
                    pass
                elif token in [LustreParser.T__5]:
                    self.state = 42
                    self.node()
                    pass
                else:
                    raise NoViableAltException(self)

                self.state = 47
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 48
            self.match(LustreParser.EOF)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DatatypedefContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.DatatypedefContext, self).__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(LustreParser.ID, 0)

        def topLevelType(self):
            return self.getTypedRuleContext(LustreParser.TopLevelTypeContext,0)


        def getRuleIndex(self):
            return LustreParser.RULE_datatypedef

        def accept(self, visitor):
            if hasattr(visitor, "visitDatatypedef"):
                return visitor.visitDatatypedef(self)
            else:
                return visitor.visitChildren(self)




    def datatypedef(self):

        localctx = LustreParser.DatatypedefContext(self, self._ctx, self.state)
        self.enterRule(localctx, 2, self.RULE_datatypedef)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 50
            self.match(LustreParser.T__0)
            self.state = 51
            self.match(LustreParser.ID)
            self.state = 52
            self.match(LustreParser.T__1)
            self.state = 53
            self.topLevelType()
            self.state = 54
            self.match(LustreParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ConstantContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.ConstantContext, self).__init__(parent, invokingState)
            self.parser = parser

        def ID(self):
            return self.getToken(LustreParser.ID, 0)

        def expr(self):
            return self.getTypedRuleContext(LustreParser.ExprContext,0)


        def datatype(self):
            return self.getTypedRuleContext(LustreParser.DatatypeContext,0)


        def getRuleIndex(self):
            return LustreParser.RULE_constant

        def accept(self, visitor):
            if hasattr(visitor, "visitConstant"):
                return visitor.visitConstant(self)
            else:
                return visitor.visitChildren(self)




    def constant(self):

        localctx = LustreParser.ConstantContext(self, self._ctx, self.state)
        self.enterRule(localctx, 4, self.RULE_constant)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 56
            self.match(LustreParser.T__3)
            self.state = 57
            self.match(LustreParser.ID)
            self.state = 60
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==LustreParser.T__4:
                self.state = 58
                self.match(LustreParser.T__4)
                self.state = 59
                self.datatype(0)


            self.state = 62
            self.match(LustreParser.T__1)
            self.state = 63
            self.expr(0)
            self.state = 64
            self.match(LustreParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class NodeContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.NodeContext, self).__init__(parent, invokingState)
            self.parser = parser
            self.inp = None # VarDeclListContext
            self.outp = None # VarDeclListContext
            self.local = None # VarDeclListContext
            self.instrlst = None # InstrListContext

        def ID(self):
            return self.getToken(LustreParser.ID, 0)

        def varDeclList(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.VarDeclListContext)
            else:
                return self.getTypedRuleContext(LustreParser.VarDeclListContext,i)


        def instrList(self):
            return self.getTypedRuleContext(LustreParser.InstrListContext,0)


        def getRuleIndex(self):
            return LustreParser.RULE_node

        def accept(self, visitor):
            if hasattr(visitor, "visitNode"):
                return visitor.visitNode(self)
            else:
                return visitor.visitChildren(self)




    def node(self):

        localctx = LustreParser.NodeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 6, self.RULE_node)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 66
            self.match(LustreParser.T__5)
            self.state = 67
            self.match(LustreParser.ID)
            self.state = 68
            self.match(LustreParser.T__6)
            self.state = 70
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==LustreParser.ID:
                self.state = 69
                localctx.inp = self.varDeclList()


            self.state = 72
            self.match(LustreParser.T__7)
            self.state = 73
            self.match(LustreParser.T__8)
            self.state = 74
            self.match(LustreParser.T__6)
            self.state = 76
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==LustreParser.ID:
                self.state = 75
                localctx.outp = self.varDeclList()


            self.state = 78
            self.match(LustreParser.T__7)
            self.state = 79
            self.match(LustreParser.T__2)
            self.state = 84
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==LustreParser.T__9:
                self.state = 80
                self.match(LustreParser.T__9)
                self.state = 81
                localctx.local = self.varDeclList()
                self.state = 82
                self.match(LustreParser.T__2)


            self.state = 86
            self.match(LustreParser.T__10)
            self.state = 88
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << LustreParser.T__6) | (1 << LustreParser.T__25) | (1 << LustreParser.T__26) | (1 << LustreParser.T__27) | (1 << LustreParser.T__28) | (1 << LustreParser.T__29) | (1 << LustreParser.ID))) != 0):
                self.state = 87
                localctx.instrlst = self.instrList()


            self.state = 90
            self.match(LustreParser.T__11)
            self.state = 92
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==LustreParser.T__2:
                self.state = 91
                self.match(LustreParser.T__2)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class InstrListContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.InstrListContext, self).__init__(parent, invokingState)
            self.parser = parser

        def instr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.InstrContext)
            else:
                return self.getTypedRuleContext(LustreParser.InstrContext,i)


        def getRuleIndex(self):
            return LustreParser.RULE_instrList

        def accept(self, visitor):
            if hasattr(visitor, "visitInstrList"):
                return visitor.visitInstrList(self)
            else:
                return visitor.visitChildren(self)




    def instrList(self):

        localctx = LustreParser.InstrListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 8, self.RULE_instrList)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 95 
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while True:
                self.state = 94
                self.instr()
                self.state = 97 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if not ((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << LustreParser.T__6) | (1 << LustreParser.T__25) | (1 << LustreParser.T__26) | (1 << LustreParser.T__27) | (1 << LustreParser.T__28) | (1 << LustreParser.T__29) | (1 << LustreParser.ID))) != 0)):
                    break

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class InstrContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.InstrContext, self).__init__(parent, invokingState)
            self.parser = parser

        def equation(self):
            return self.getTypedRuleContext(LustreParser.EquationContext,0)


        def prop(self):
            return self.getTypedRuleContext(LustreParser.PropContext,0)


        def assertion(self):
            return self.getTypedRuleContext(LustreParser.AssertionContext,0)


        def main(self):
            return self.getTypedRuleContext(LustreParser.MainContext,0)


        def realizabilityInputs(self):
            return self.getTypedRuleContext(LustreParser.RealizabilityInputsContext,0)


        def ivc(self):
            return self.getTypedRuleContext(LustreParser.IvcContext,0)


        def getRuleIndex(self):
            return LustreParser.RULE_instr

        def accept(self, visitor):
            if hasattr(visitor, "visitInstr"):
                return visitor.visitInstr(self)
            else:
                return visitor.visitChildren(self)




    def instr(self):

        localctx = LustreParser.InstrContext(self, self._ctx, self.state)
        self.enterRule(localctx, 10, self.RULE_instr)
        try:
            self.state = 105
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [LustreParser.T__6, LustreParser.ID]:
                self.enterOuterAlt(localctx, 1)
                self.state = 99
                self.equation()
                pass
            elif token in [LustreParser.T__25]:
                self.enterOuterAlt(localctx, 2)
                self.state = 100
                self.prop()
                pass
            elif token in [LustreParser.T__29]:
                self.enterOuterAlt(localctx, 3)
                self.state = 101
                self.assertion()
                pass
            elif token in [LustreParser.T__28]:
                self.enterOuterAlt(localctx, 4)
                self.state = 102
                self.main()
                pass
            elif token in [LustreParser.T__26]:
                self.enterOuterAlt(localctx, 5)
                self.state = 103
                self.realizabilityInputs()
                pass
            elif token in [LustreParser.T__27]:
                self.enterOuterAlt(localctx, 6)
                self.state = 104
                self.ivc()
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class VarDeclListContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.VarDeclListContext, self).__init__(parent, invokingState)
            self.parser = parser

        def varDeclGroup(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.VarDeclGroupContext)
            else:
                return self.getTypedRuleContext(LustreParser.VarDeclGroupContext,i)


        def getRuleIndex(self):
            return LustreParser.RULE_varDeclList

        def accept(self, visitor):
            if hasattr(visitor, "visitVarDeclList"):
                return visitor.visitVarDeclList(self)
            else:
                return visitor.visitChildren(self)




    def varDeclList(self):

        localctx = LustreParser.VarDeclListContext(self, self._ctx, self.state)
        self.enterRule(localctx, 12, self.RULE_varDeclList)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 107
            self.varDeclGroup()
            self.state = 112
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,10,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    self.state = 108
                    self.match(LustreParser.T__2)
                    self.state = 109
                    self.varDeclGroup() 
                self.state = 114
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,10,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class VarDeclGroupContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.VarDeclGroupContext, self).__init__(parent, invokingState)
            self.parser = parser

        def eID(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.EIDContext)
            else:
                return self.getTypedRuleContext(LustreParser.EIDContext,i)


        def datatype(self):
            return self.getTypedRuleContext(LustreParser.DatatypeContext,0)


        def getRuleIndex(self):
            return LustreParser.RULE_varDeclGroup

        def accept(self, visitor):
            if hasattr(visitor, "visitVarDeclGroup"):
                return visitor.visitVarDeclGroup(self)
            else:
                return visitor.visitChildren(self)




    def varDeclGroup(self):

        localctx = LustreParser.VarDeclGroupContext(self, self._ctx, self.state)
        self.enterRule(localctx, 14, self.RULE_varDeclGroup)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 115
            self.eID(0)
            self.state = 120
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==LustreParser.T__12:
                self.state = 116
                self.match(LustreParser.T__12)
                self.state = 117
                self.eID(0)
                self.state = 122
                self._errHandler.sync(self)
                _la = self._input.LA(1)

            self.state = 123
            self.match(LustreParser.T__4)
            self.state = 124
            self.datatype(0)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class TopLevelTypeContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.TopLevelTypeContext, self).__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return LustreParser.RULE_topLevelType

     
        def copyFrom(self, ctx):
            super(LustreParser.TopLevelTypeContext, self).copyFrom(ctx)



    class RecordTypeContext(TopLevelTypeContext):

        def __init__(self, parser, ctx): # actually a LustreParser.TopLevelTypeContext)
            super(LustreParser.RecordTypeContext, self).__init__(parser)
            self.copyFrom(ctx)

        def ID(self, i=None):
            if i is None:
                return self.getTokens(LustreParser.ID)
            else:
                return self.getToken(LustreParser.ID, i)
        def datatype(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.DatatypeContext)
            else:
                return self.getTypedRuleContext(LustreParser.DatatypeContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitRecordType"):
                return visitor.visitRecordType(self)
            else:
                return visitor.visitChildren(self)


    class EnumTypeContext(TopLevelTypeContext):

        def __init__(self, parser, ctx): # actually a LustreParser.TopLevelTypeContext)
            super(LustreParser.EnumTypeContext, self).__init__(parser)
            self.copyFrom(ctx)

        def ID(self, i=None):
            if i is None:
                return self.getTokens(LustreParser.ID)
            else:
                return self.getToken(LustreParser.ID, i)

        def accept(self, visitor):
            if hasattr(visitor, "visitEnumType"):
                return visitor.visitEnumType(self)
            else:
                return visitor.visitChildren(self)


    class PlainTypeContext(TopLevelTypeContext):

        def __init__(self, parser, ctx): # actually a LustreParser.TopLevelTypeContext)
            super(LustreParser.PlainTypeContext, self).__init__(parser)
            self.copyFrom(ctx)

        def datatype(self):
            return self.getTypedRuleContext(LustreParser.DatatypeContext,0)


        def accept(self, visitor):
            if hasattr(visitor, "visitPlainType"):
                return visitor.visitPlainType(self)
            else:
                return visitor.visitChildren(self)



    def topLevelType(self):

        localctx = LustreParser.TopLevelTypeContext(self, self._ctx, self.state)
        self.enterRule(localctx, 16, self.RULE_topLevelType)
        self._la = 0 # Token type
        try:
            self.state = 155
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [LustreParser.T__17, LustreParser.T__18, LustreParser.T__22, LustreParser.T__23, LustreParser.ID]:
                localctx = LustreParser.PlainTypeContext(self, localctx)
                self.enterOuterAlt(localctx, 1)
                self.state = 126
                self.datatype(0)
                pass
            elif token in [LustreParser.T__13]:
                localctx = LustreParser.RecordTypeContext(self, localctx)
                self.enterOuterAlt(localctx, 2)
                self.state = 127
                self.match(LustreParser.T__13)
                self.state = 128
                self.match(LustreParser.T__14)

                self.state = 129
                self.match(LustreParser.ID)
                self.state = 130
                self.match(LustreParser.T__4)
                self.state = 131
                self.datatype(0)
                self.state = 139
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==LustreParser.T__2:
                    self.state = 133
                    self.match(LustreParser.T__2)
                    self.state = 134
                    self.match(LustreParser.ID)
                    self.state = 135
                    self.match(LustreParser.T__4)
                    self.state = 136
                    self.datatype(0)
                    self.state = 141
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 142
                self.match(LustreParser.T__15)
                pass
            elif token in [LustreParser.T__16]:
                localctx = LustreParser.EnumTypeContext(self, localctx)
                self.enterOuterAlt(localctx, 3)
                self.state = 144
                self.match(LustreParser.T__16)
                self.state = 145
                self.match(LustreParser.T__14)
                self.state = 146
                self.match(LustreParser.ID)
                self.state = 151
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==LustreParser.T__12:
                    self.state = 147
                    self.match(LustreParser.T__12)
                    self.state = 148
                    self.match(LustreParser.ID)
                    self.state = 153
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 154
                self.match(LustreParser.T__15)
                pass
            else:
                raise NoViableAltException(self)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class DatatypeContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.DatatypeContext, self).__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return LustreParser.RULE_datatype

     
        def copyFrom(self, ctx):
            super(LustreParser.DatatypeContext, self).copyFrom(ctx)


    class ArrayTypeContext(DatatypeContext):

        def __init__(self, parser, ctx): # actually a LustreParser.DatatypeContext)
            super(LustreParser.ArrayTypeContext, self).__init__(parser)
            self.copyFrom(ctx)

        def datatype(self):
            return self.getTypedRuleContext(LustreParser.DatatypeContext,0)

        def INT(self):
            return self.getToken(LustreParser.INT, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitArrayType"):
                return visitor.visitArrayType(self)
            else:
                return visitor.visitChildren(self)


    class SubrangeTypeContext(DatatypeContext):

        def __init__(self, parser, ctx): # actually a LustreParser.DatatypeContext)
            super(LustreParser.SubrangeTypeContext, self).__init__(parser)
            self.copyFrom(ctx)

        def bound(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.BoundContext)
            else:
                return self.getTypedRuleContext(LustreParser.BoundContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitSubrangeType"):
                return visitor.visitSubrangeType(self)
            else:
                return visitor.visitChildren(self)


    class UserTypeContext(DatatypeContext):

        def __init__(self, parser, ctx): # actually a LustreParser.DatatypeContext)
            super(LustreParser.UserTypeContext, self).__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(LustreParser.ID, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitUserType"):
                return visitor.visitUserType(self)
            else:
                return visitor.visitChildren(self)


    class PrimitiveTypeContext(DatatypeContext):

        def __init__(self, parser, ctx): # actually a LustreParser.DatatypeContext)
            super(LustreParser.PrimitiveTypeContext, self).__init__(parser)
            self.copyFrom(ctx)


        def accept(self, visitor):
            if hasattr(visitor, "visitPrimitiveType"):
                return visitor.visitPrimitiveType(self)
            else:
                return visitor.visitChildren(self)



    def datatype(self, _p=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = LustreParser.DatatypeContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 18
        self.enterRecursionRule(localctx, 18, self.RULE_datatype, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 171
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [LustreParser.T__17]:
                localctx = LustreParser.PrimitiveTypeContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 158
                self.match(LustreParser.T__17)
                pass
            elif token in [LustreParser.T__18]:
                localctx = LustreParser.SubrangeTypeContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 159
                self.match(LustreParser.T__18)
                self.state = 160
                self.match(LustreParser.T__19)
                self.state = 161
                self.bound()
                self.state = 162
                self.match(LustreParser.T__12)
                self.state = 163
                self.bound()
                self.state = 164
                self.match(LustreParser.T__20)
                self.state = 165
                self.match(LustreParser.T__21)
                self.state = 166
                self.match(LustreParser.T__17)
                pass
            elif token in [LustreParser.T__22]:
                localctx = LustreParser.PrimitiveTypeContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 168
                self.match(LustreParser.T__22)
                pass
            elif token in [LustreParser.T__23]:
                localctx = LustreParser.PrimitiveTypeContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 169
                self.match(LustreParser.T__23)
                pass
            elif token in [LustreParser.ID]:
                localctx = LustreParser.UserTypeContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 170
                self.match(LustreParser.ID)
                pass
            else:
                raise NoViableAltException(self)

            self._ctx.stop = self._input.LT(-1)
            self.state = 179
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,16,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    localctx = LustreParser.ArrayTypeContext(self, LustreParser.DatatypeContext(self, _parentctx, _parentState))
                    self.pushNewRecursionContext(localctx, _startState, self.RULE_datatype)
                    self.state = 173
                    if not self.precpred(self._ctx, 2):
                        from antlr4.error.Errors import FailedPredicateException
                        raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                    self.state = 174
                    self.match(LustreParser.T__19)
                    self.state = 175
                    self.match(LustreParser.INT)
                    self.state = 176
                    self.match(LustreParser.T__20) 
                self.state = 181
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,16,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class BoundContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.BoundContext, self).__init__(parent, invokingState)
            self.parser = parser

        def INT(self):
            return self.getToken(LustreParser.INT, 0)

        def getRuleIndex(self):
            return LustreParser.RULE_bound

        def accept(self, visitor):
            if hasattr(visitor, "visitBound"):
                return visitor.visitBound(self)
            else:
                return visitor.visitChildren(self)




    def bound(self):

        localctx = LustreParser.BoundContext(self, self._ctx, self.state)
        self.enterRule(localctx, 20, self.RULE_bound)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 183
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==LustreParser.T__24:
                self.state = 182
                self.match(LustreParser.T__24)


            self.state = 185
            self.match(LustreParser.INT)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class PropContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.PropContext, self).__init__(parent, invokingState)
            self.parser = parser

        def expr(self):
            return self.getTypedRuleContext(LustreParser.ExprContext,0)


        def getRuleIndex(self):
            return LustreParser.RULE_prop

        def accept(self, visitor):
            if hasattr(visitor, "visitProp"):
                return visitor.visitProp(self)
            else:
                return visitor.visitChildren(self)




    def prop(self):

        localctx = LustreParser.PropContext(self, self._ctx, self.state)
        self.enterRule(localctx, 22, self.RULE_prop)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 187
            self.match(LustreParser.T__25)
            self.state = 188
            self.expr(0)
            self.state = 189
            self.match(LustreParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class RealizabilityInputsContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.RealizabilityInputsContext, self).__init__(parent, invokingState)
            self.parser = parser

        def ID(self, i=None):
            if i is None:
                return self.getTokens(LustreParser.ID)
            else:
                return self.getToken(LustreParser.ID, i)

        def getRuleIndex(self):
            return LustreParser.RULE_realizabilityInputs

        def accept(self, visitor):
            if hasattr(visitor, "visitRealizabilityInputs"):
                return visitor.visitRealizabilityInputs(self)
            else:
                return visitor.visitChildren(self)




    def realizabilityInputs(self):

        localctx = LustreParser.RealizabilityInputsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 24, self.RULE_realizabilityInputs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 191
            self.match(LustreParser.T__26)
            self.state = 200
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==LustreParser.ID:
                self.state = 192
                self.match(LustreParser.ID)
                self.state = 197
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==LustreParser.T__12:
                    self.state = 193
                    self.match(LustreParser.T__12)
                    self.state = 194
                    self.match(LustreParser.ID)
                    self.state = 199
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)



            self.state = 202
            self.match(LustreParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class IvcContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.IvcContext, self).__init__(parent, invokingState)
            self.parser = parser

        def ID(self, i=None):
            if i is None:
                return self.getTokens(LustreParser.ID)
            else:
                return self.getToken(LustreParser.ID, i)

        def getRuleIndex(self):
            return LustreParser.RULE_ivc

        def accept(self, visitor):
            if hasattr(visitor, "visitIvc"):
                return visitor.visitIvc(self)
            else:
                return visitor.visitChildren(self)




    def ivc(self):

        localctx = LustreParser.IvcContext(self, self._ctx, self.state)
        self.enterRule(localctx, 26, self.RULE_ivc)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 204
            self.match(LustreParser.T__27)
            self.state = 213
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==LustreParser.ID:
                self.state = 205
                self.match(LustreParser.ID)
                self.state = 210
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==LustreParser.T__12:
                    self.state = 206
                    self.match(LustreParser.T__12)
                    self.state = 207
                    self.match(LustreParser.ID)
                    self.state = 212
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)



            self.state = 215
            self.match(LustreParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class MainContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.MainContext, self).__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return LustreParser.RULE_main

        def accept(self, visitor):
            if hasattr(visitor, "visitMain"):
                return visitor.visitMain(self)
            else:
                return visitor.visitChildren(self)




    def main(self):

        localctx = LustreParser.MainContext(self, self._ctx, self.state)
        self.enterRule(localctx, 28, self.RULE_main)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 217
            self.match(LustreParser.T__28)
            self.state = 219
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            if _la==LustreParser.T__2:
                self.state = 218
                self.match(LustreParser.T__2)


        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class AssertionContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.AssertionContext, self).__init__(parent, invokingState)
            self.parser = parser

        def expr(self):
            return self.getTypedRuleContext(LustreParser.ExprContext,0)


        def getRuleIndex(self):
            return LustreParser.RULE_assertion

        def accept(self, visitor):
            if hasattr(visitor, "visitAssertion"):
                return visitor.visitAssertion(self)
            else:
                return visitor.visitChildren(self)




    def assertion(self):

        localctx = LustreParser.AssertionContext(self, self._ctx, self.state)
        self.enterRule(localctx, 30, self.RULE_assertion)
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 221
            self.match(LustreParser.T__29)
            self.state = 222
            self.expr(0)
            self.state = 223
            self.match(LustreParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class EquationContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.EquationContext, self).__init__(parent, invokingState)
            self.parser = parser
            self.pos1 = None # LhsContext
            self.pos2 = None # LhsContext
            self.pos3 = None # ExprContext

        def expr(self):
            return self.getTypedRuleContext(LustreParser.ExprContext,0)


        def lhs(self):
            return self.getTypedRuleContext(LustreParser.LhsContext,0)


        def getRuleIndex(self):
            return LustreParser.RULE_equation

        def accept(self, visitor):
            if hasattr(visitor, "visitEquation"):
                return visitor.visitEquation(self)
            else:
                return visitor.visitChildren(self)




    def equation(self):

        localctx = LustreParser.EquationContext(self, self._ctx, self.state)
        self.enterRule(localctx, 32, self.RULE_equation)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 231
            self._errHandler.sync(self)
            token = self._input.LA(1)
            if token in [LustreParser.ID]:
                self.state = 225
                localctx.pos1 = self.lhs()
                pass
            elif token in [LustreParser.T__6]:
                self.state = 226
                self.match(LustreParser.T__6)
                self.state = 228
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==LustreParser.ID:
                    self.state = 227
                    localctx.pos2 = self.lhs()


                self.state = 230
                self.match(LustreParser.T__7)
                pass
            else:
                raise NoViableAltException(self)

            self.state = 233
            self.match(LustreParser.T__1)
            self.state = 234
            localctx.pos3 = self.expr(0)
            self.state = 235
            self.match(LustreParser.T__2)
        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class LhsContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.LhsContext, self).__init__(parent, invokingState)
            self.parser = parser

        def eID(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.EIDContext)
            else:
                return self.getTypedRuleContext(LustreParser.EIDContext,i)


        def getRuleIndex(self):
            return LustreParser.RULE_lhs

        def accept(self, visitor):
            if hasattr(visitor, "visitLhs"):
                return visitor.visitLhs(self)
            else:
                return visitor.visitChildren(self)




    def lhs(self):

        localctx = LustreParser.LhsContext(self, self._ctx, self.state)
        self.enterRule(localctx, 34, self.RULE_lhs)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 237
            self.eID(0)
            self.state = 242
            self._errHandler.sync(self)
            _la = self._input.LA(1)
            while _la==LustreParser.T__12:
                self.state = 238
                self.match(LustreParser.T__12)
                self.state = 239
                self.eID(0)
                self.state = 244
                self._errHandler.sync(self)
                _la = self._input.LA(1)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.exitRule()
        return localctx

    class ExprContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.ExprContext, self).__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return LustreParser.RULE_expr

     
        def copyFrom(self, ctx):
            super(LustreParser.ExprContext, self).copyFrom(ctx)


    class RecordExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.RecordExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def ID(self, i=None):
            if i is None:
                return self.getTokens(LustreParser.ID)
            else:
                return self.getToken(LustreParser.ID, i)
        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitRecordExpr"):
                return visitor.visitRecordExpr(self)
            else:
                return visitor.visitChildren(self)


    class IntExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.IntExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def INT(self):
            return self.getToken(LustreParser.INT, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitIntExpr"):
                return visitor.visitIntExpr(self)
            else:
                return visitor.visitChildren(self)


    class ArrayExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.ArrayExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitArrayExpr"):
                return visitor.visitArrayExpr(self)
            else:
                return visitor.visitChildren(self)


    class CastExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.CastExprContext, self).__init__(parser)
            self.op = None # Token
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(LustreParser.ExprContext,0)


        def accept(self, visitor):
            if hasattr(visitor, "visitCastExpr"):
                return visitor.visitCastExpr(self)
            else:
                return visitor.visitChildren(self)


    class RealExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.RealExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def REAL(self):
            return self.getToken(LustreParser.REAL, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitRealExpr"):
                return visitor.visitRealExpr(self)
            else:
                return visitor.visitChildren(self)


    class IfThenElseExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.IfThenElseExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitIfThenElseExpr"):
                return visitor.visitIfThenElseExpr(self)
            else:
                return visitor.visitChildren(self)


    class BinaryExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.BinaryExprContext, self).__init__(parser)
            self.op = None # Token
            self.copyFrom(ctx)

        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitBinaryExpr"):
                return visitor.visitBinaryExpr(self)
            else:
                return visitor.visitChildren(self)


    class PreExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.PreExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(LustreParser.ExprContext,0)


        def accept(self, visitor):
            if hasattr(visitor, "visitPreExpr"):
                return visitor.visitPreExpr(self)
            else:
                return visitor.visitChildren(self)


    class InitCurrExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.InitCurrExprContext, self).__init__(parser)
            self.op = None # Token
            self.copyFrom(ctx)

        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitInitCurrExpr"):
                return visitor.visitInitCurrExpr(self)
            else:
                return visitor.visitChildren(self)


    class NodeCallExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.NodeCallExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(LustreParser.ID, 0)
        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitNodeCallExpr"):
                return visitor.visitNodeCallExpr(self)
            else:
                return visitor.visitChildren(self)


    class RecordAccessExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.RecordAccessExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(LustreParser.ExprContext,0)

        def ID(self):
            return self.getToken(LustreParser.ID, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitRecordAccessExpr"):
                return visitor.visitRecordAccessExpr(self)
            else:
                return visitor.visitChildren(self)


    class NegateExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.NegateExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(LustreParser.ExprContext,0)


        def accept(self, visitor):
            if hasattr(visitor, "visitNegateExpr"):
                return visitor.visitNegateExpr(self)
            else:
                return visitor.visitChildren(self)


    class CondactExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.CondactExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitCondactExpr"):
                return visitor.visitCondactExpr(self)
            else:
                return visitor.visitChildren(self)


    class NotExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.NotExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self):
            return self.getTypedRuleContext(LustreParser.ExprContext,0)


        def accept(self, visitor):
            if hasattr(visitor, "visitNotExpr"):
                return visitor.visitNotExpr(self)
            else:
                return visitor.visitChildren(self)


    class ArrayAccessExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.ArrayAccessExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitArrayAccessExpr"):
                return visitor.visitArrayAccessExpr(self)
            else:
                return visitor.visitChildren(self)


    class ArrayUpdateExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.ArrayUpdateExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitArrayUpdateExpr"):
                return visitor.visitArrayUpdateExpr(self)
            else:
                return visitor.visitChildren(self)


    class BoolExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.BoolExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def BOOL(self):
            return self.getToken(LustreParser.BOOL, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitBoolExpr"):
                return visitor.visitBoolExpr(self)
            else:
                return visitor.visitChildren(self)


    class TupleExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.TupleExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)


        def accept(self, visitor):
            if hasattr(visitor, "visitTupleExpr"):
                return visitor.visitTupleExpr(self)
            else:
                return visitor.visitChildren(self)


    class RecordUpdateExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.RecordUpdateExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def expr(self, i=None):
            if i is None:
                return self.getTypedRuleContexts(LustreParser.ExprContext)
            else:
                return self.getTypedRuleContext(LustreParser.ExprContext,i)

        def ID(self):
            return self.getToken(LustreParser.ID, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitRecordUpdateExpr"):
                return visitor.visitRecordUpdateExpr(self)
            else:
                return visitor.visitChildren(self)


    class IdExprContext(ExprContext):

        def __init__(self, parser, ctx): # actually a LustreParser.ExprContext)
            super(LustreParser.IdExprContext, self).__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(LustreParser.ID, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitIdExpr"):
                return visitor.visitIdExpr(self)
            else:
                return visitor.visitChildren(self)



    def expr(self, _p=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = LustreParser.ExprContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 36
        self.enterRecursionRule(localctx, 36, self.RULE_expr, _p)
        self._la = 0 # Token type
        try:
            self.enterOuterAlt(localctx, 1)
            self.state = 333
            self._errHandler.sync(self)
            la_ = self._interp.adaptivePredict(self._input,33,self._ctx)
            if la_ == 1:
                localctx = LustreParser.IdExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx

                self.state = 246
                self.match(LustreParser.ID)
                pass

            elif la_ == 2:
                localctx = LustreParser.IntExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 248
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if _la==LustreParser.T__24:
                    self.state = 247
                    self.match(LustreParser.T__24)


                self.state = 250
                self.match(LustreParser.INT)
                pass

            elif la_ == 3:
                localctx = LustreParser.RealExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 251
                self.match(LustreParser.REAL)
                pass

            elif la_ == 4:
                localctx = LustreParser.BoolExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 252
                self.match(LustreParser.BOOL)
                pass

            elif la_ == 5:
                localctx = LustreParser.CastExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 253
                localctx.op = self._input.LT(1)
                _la = self._input.LA(1)
                if not(_la==LustreParser.T__23 or _la==LustreParser.T__30):
                    localctx.op = self._errHandler.recoverInline(self)
                else:
                    self._errHandler.reportMatch(self)
                    self.consume()
                self.state = 254
                self.match(LustreParser.T__6)
                self.state = 255
                self.expr(0)
                self.state = 256
                self.match(LustreParser.T__7)
                pass

            elif la_ == 6:
                localctx = LustreParser.NodeCallExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 258
                self.match(LustreParser.ID)
                self.state = 259
                self.match(LustreParser.T__6)
                self.state = 268
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                if (((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << LustreParser.T__6) | (1 << LustreParser.T__19) | (1 << LustreParser.T__23) | (1 << LustreParser.T__24) | (1 << LustreParser.T__30) | (1 << LustreParser.T__31) | (1 << LustreParser.T__34) | (1 << LustreParser.T__35) | (1 << LustreParser.T__51) | (1 << LustreParser.REAL) | (1 << LustreParser.BOOL) | (1 << LustreParser.INT) | (1 << LustreParser.ID))) != 0):
                    self.state = 260
                    self.expr(0)
                    self.state = 265
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    while _la==LustreParser.T__12:
                        self.state = 261
                        self.match(LustreParser.T__12)
                        self.state = 262
                        self.expr(0)
                        self.state = 267
                        self._errHandler.sync(self)
                        _la = self._input.LA(1)



                self.state = 270
                self.match(LustreParser.T__7)
                pass

            elif la_ == 7:
                localctx = LustreParser.CondactExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 271
                self.match(LustreParser.T__31)
                self.state = 272
                self.match(LustreParser.T__6)
                self.state = 273
                self.expr(0)
                self.state = 276 
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while True:
                    self.state = 274
                    self.match(LustreParser.T__12)
                    self.state = 275
                    self.expr(0)
                    self.state = 278 
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)
                    if not (_la==LustreParser.T__12):
                        break

                self.state = 280
                self.match(LustreParser.T__7)
                pass

            elif la_ == 8:
                localctx = LustreParser.PreExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 282
                self.match(LustreParser.T__34)
                self.state = 283
                self.expr(14)
                pass

            elif la_ == 9:
                localctx = LustreParser.NotExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 284
                self.match(LustreParser.T__35)
                self.state = 285
                self.expr(13)
                pass

            elif la_ == 10:
                localctx = LustreParser.NegateExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 286
                self.match(LustreParser.T__24)
                self.state = 287
                self.expr(12)
                pass

            elif la_ == 11:
                localctx = LustreParser.IfThenElseExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 288
                self.match(LustreParser.T__51)
                self.state = 289
                self.expr(0)
                self.state = 290
                self.match(LustreParser.T__52)
                self.state = 291
                self.expr(0)
                self.state = 292
                self.match(LustreParser.T__53)
                self.state = 293
                self.expr(4)
                pass

            elif la_ == 12:
                localctx = LustreParser.RecordExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 295
                self.match(LustreParser.ID)
                self.state = 296
                self.match(LustreParser.T__14)
                self.state = 297
                self.match(LustreParser.ID)
                self.state = 298
                self.match(LustreParser.T__1)
                self.state = 299
                self.expr(0)
                self.state = 306
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==LustreParser.T__2:
                    self.state = 300
                    self.match(LustreParser.T__2)
                    self.state = 301
                    self.match(LustreParser.ID)
                    self.state = 302
                    self.match(LustreParser.T__1)
                    self.state = 303
                    self.expr(0)
                    self.state = 308
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 309
                self.match(LustreParser.T__15)
                pass

            elif la_ == 13:
                localctx = LustreParser.ArrayExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 311
                self.match(LustreParser.T__19)
                self.state = 312
                self.expr(0)
                self.state = 317
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==LustreParser.T__12:
                    self.state = 313
                    self.match(LustreParser.T__12)
                    self.state = 314
                    self.expr(0)
                    self.state = 319
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 320
                self.match(LustreParser.T__20)
                pass

            elif la_ == 14:
                localctx = LustreParser.TupleExprContext(self, localctx)
                self._ctx = localctx
                _prevctx = localctx
                self.state = 322
                self.match(LustreParser.T__6)
                self.state = 323
                self.expr(0)
                self.state = 328
                self._errHandler.sync(self)
                _la = self._input.LA(1)
                while _la==LustreParser.T__12:
                    self.state = 324
                    self.match(LustreParser.T__12)
                    self.state = 325
                    self.expr(0)
                    self.state = 330
                    self._errHandler.sync(self)
                    _la = self._input.LA(1)

                self.state = 331
                self.match(LustreParser.T__7)
                pass


            self._ctx.stop = self._input.LT(-1)
            self.state = 380
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,35,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 378
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,34,self._ctx)
                    if la_ == 1:
                        localctx = LustreParser.BinaryExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 335
                        if not self.precpred(self._ctx, 11):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 11)")
                        self.state = 336
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << LustreParser.T__36) | (1 << LustreParser.T__37) | (1 << LustreParser.T__38) | (1 << LustreParser.T__39))) != 0)):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 337
                        self.expr(12)
                        pass

                    elif la_ == 2:
                        localctx = LustreParser.BinaryExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 338
                        if not self.precpred(self._ctx, 10):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 10)")
                        self.state = 339
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not(_la==LustreParser.T__24 or _la==LustreParser.T__40):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 340
                        self.expr(11)
                        pass

                    elif la_ == 3:
                        localctx = LustreParser.BinaryExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 341
                        if not self.precpred(self._ctx, 9):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 9)")
                        self.state = 342
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not((((_la) & ~0x3f) == 0 and ((1 << _la) & ((1 << LustreParser.T__1) | (1 << LustreParser.T__41) | (1 << LustreParser.T__42) | (1 << LustreParser.T__43) | (1 << LustreParser.T__44) | (1 << LustreParser.T__45))) != 0)):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 343
                        self.expr(10)
                        pass

                    elif la_ == 4:
                        localctx = LustreParser.BinaryExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 344
                        if not self.precpred(self._ctx, 8):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 8)")
                        self.state = 345
                        localctx.op = self.match(LustreParser.T__46)
                        self.state = 346
                        self.expr(9)
                        pass

                    elif la_ == 5:
                        localctx = LustreParser.BinaryExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 347
                        if not self.precpred(self._ctx, 7):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 7)")
                        self.state = 348
                        localctx.op = self._input.LT(1)
                        _la = self._input.LA(1)
                        if not(_la==LustreParser.T__47 or _la==LustreParser.T__48):
                            localctx.op = self._errHandler.recoverInline(self)
                        else:
                            self._errHandler.reportMatch(self)
                            self.consume()
                        self.state = 349
                        self.expr(8)
                        pass

                    elif la_ == 6:
                        localctx = LustreParser.BinaryExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 350
                        if not self.precpred(self._ctx, 6):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 6)")
                        self.state = 351
                        localctx.op = self.match(LustreParser.T__49)
                        self.state = 352
                        self.expr(6)
                        pass

                    elif la_ == 7:
                        localctx = LustreParser.InitCurrExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 353
                        if not self.precpred(self._ctx, 5):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 5)")
                        self.state = 354
                        localctx.op = self.match(LustreParser.T__50)
                        self.state = 355
                        self.expr(5)
                        pass

                    elif la_ == 8:
                        localctx = LustreParser.RecordAccessExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 356
                        if not self.precpred(self._ctx, 18):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 18)")
                        self.state = 357
                        self.match(LustreParser.T__32)
                        self.state = 358
                        self.match(LustreParser.ID)
                        pass

                    elif la_ == 9:
                        localctx = LustreParser.RecordUpdateExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 359
                        if not self.precpred(self._ctx, 17):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 17)")
                        self.state = 360
                        self.match(LustreParser.T__14)
                        self.state = 361
                        self.match(LustreParser.ID)
                        self.state = 362
                        self.match(LustreParser.T__33)
                        self.state = 363
                        self.expr(0)
                        self.state = 364
                        self.match(LustreParser.T__15)
                        pass

                    elif la_ == 10:
                        localctx = LustreParser.ArrayAccessExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 366
                        if not self.precpred(self._ctx, 16):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 16)")
                        self.state = 367
                        self.match(LustreParser.T__19)
                        self.state = 368
                        self.expr(0)
                        self.state = 369
                        self.match(LustreParser.T__20)
                        pass

                    elif la_ == 11:
                        localctx = LustreParser.ArrayUpdateExprContext(self, LustreParser.ExprContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_expr)
                        self.state = 371
                        if not self.precpred(self._ctx, 15):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 15)")
                        self.state = 372
                        self.match(LustreParser.T__19)
                        self.state = 373
                        self.expr(0)
                        self.state = 374
                        self.match(LustreParser.T__33)
                        self.state = 375
                        self.expr(0)
                        self.state = 376
                        self.match(LustreParser.T__20)
                        pass

             
                self.state = 382
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,35,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx

    class EIDContext(ParserRuleContext):

        def __init__(self, parser, parent=None, invokingState=-1):
            super(LustreParser.EIDContext, self).__init__(parent, invokingState)
            self.parser = parser


        def getRuleIndex(self):
            return LustreParser.RULE_eID

     
        def copyFrom(self, ctx):
            super(LustreParser.EIDContext, self).copyFrom(ctx)


    class BaseEIDContext(EIDContext):

        def __init__(self, parser, ctx): # actually a LustreParser.EIDContext)
            super(LustreParser.BaseEIDContext, self).__init__(parser)
            self.copyFrom(ctx)

        def ID(self):
            return self.getToken(LustreParser.ID, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitBaseEID"):
                return visitor.visitBaseEID(self)
            else:
                return visitor.visitChildren(self)


    class ArrayEIDContext(EIDContext):

        def __init__(self, parser, ctx): # actually a LustreParser.EIDContext)
            super(LustreParser.ArrayEIDContext, self).__init__(parser)
            self.copyFrom(ctx)

        def eID(self):
            return self.getTypedRuleContext(LustreParser.EIDContext,0)

        def INT(self):
            return self.getToken(LustreParser.INT, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitArrayEID"):
                return visitor.visitArrayEID(self)
            else:
                return visitor.visitChildren(self)


    class RecordEIDContext(EIDContext):

        def __init__(self, parser, ctx): # actually a LustreParser.EIDContext)
            super(LustreParser.RecordEIDContext, self).__init__(parser)
            self.copyFrom(ctx)

        def eID(self):
            return self.getTypedRuleContext(LustreParser.EIDContext,0)

        def ID(self):
            return self.getToken(LustreParser.ID, 0)

        def accept(self, visitor):
            if hasattr(visitor, "visitRecordEID"):
                return visitor.visitRecordEID(self)
            else:
                return visitor.visitChildren(self)



    def eID(self, _p=0):
        _parentctx = self._ctx
        _parentState = self.state
        localctx = LustreParser.EIDContext(self, self._ctx, _parentState)
        _prevctx = localctx
        _startState = 38
        self.enterRecursionRule(localctx, 38, self.RULE_eID, _p)
        try:
            self.enterOuterAlt(localctx, 1)
            localctx = LustreParser.BaseEIDContext(self, localctx)
            self._ctx = localctx
            _prevctx = localctx

            self.state = 384
            self.match(LustreParser.ID)
            self._ctx.stop = self._input.LT(-1)
            self.state = 395
            self._errHandler.sync(self)
            _alt = self._interp.adaptivePredict(self._input,37,self._ctx)
            while _alt!=2 and _alt!=ATN.INVALID_ALT_NUMBER:
                if _alt==1:
                    if self._parseListeners is not None:
                        self.triggerExitRuleEvent()
                    _prevctx = localctx
                    self.state = 393
                    self._errHandler.sync(self)
                    la_ = self._interp.adaptivePredict(self._input,36,self._ctx)
                    if la_ == 1:
                        localctx = LustreParser.ArrayEIDContext(self, LustreParser.EIDContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_eID)
                        self.state = 386
                        if not self.precpred(self._ctx, 2):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 2)")
                        self.state = 387
                        self.match(LustreParser.T__19)
                        self.state = 388
                        self.match(LustreParser.INT)
                        self.state = 389
                        self.match(LustreParser.T__20)
                        pass

                    elif la_ == 2:
                        localctx = LustreParser.RecordEIDContext(self, LustreParser.EIDContext(self, _parentctx, _parentState))
                        self.pushNewRecursionContext(localctx, _startState, self.RULE_eID)
                        self.state = 390
                        if not self.precpred(self._ctx, 1):
                            from antlr4.error.Errors import FailedPredicateException
                            raise FailedPredicateException(self, "self.precpred(self._ctx, 1)")
                        self.state = 391
                        self.match(LustreParser.T__32)
                        self.state = 392
                        self.match(LustreParser.ID)
                        pass

             
                self.state = 397
                self._errHandler.sync(self)
                _alt = self._interp.adaptivePredict(self._input,37,self._ctx)

        except RecognitionException as re:
            localctx.exception = re
            self._errHandler.reportError(self, re)
            self._errHandler.recover(self, re)
        finally:
            self.unrollRecursionContexts(_parentctx)
        return localctx



    def sempred(self, localctx, ruleIndex, predIndex):
        if self._predicates == None:
            self._predicates = dict()
        self._predicates[9] = self.datatype_sempred
        self._predicates[18] = self.expr_sempred
        self._predicates[19] = self.eID_sempred
        pred = self._predicates.get(ruleIndex, None)
        if pred is None:
            raise Exception("No predicate with index:" + str(ruleIndex))
        else:
            return pred(localctx, predIndex)

    def datatype_sempred(self, localctx, predIndex):
            if predIndex == 0:
                return self.precpred(self._ctx, 2)
         

    def expr_sempred(self, localctx, predIndex):
            if predIndex == 1:
                return self.precpred(self._ctx, 11)
         

            if predIndex == 2:
                return self.precpred(self._ctx, 10)
         

            if predIndex == 3:
                return self.precpred(self._ctx, 9)
         

            if predIndex == 4:
                return self.precpred(self._ctx, 8)
         

            if predIndex == 5:
                return self.precpred(self._ctx, 7)
         

            if predIndex == 6:
                return self.precpred(self._ctx, 6)
         

            if predIndex == 7:
                return self.precpred(self._ctx, 5)
         

            if predIndex == 8:
                return self.precpred(self._ctx, 18)
         

            if predIndex == 9:
                return self.precpred(self._ctx, 17)
         

            if predIndex == 10:
                return self.precpred(self._ctx, 16)
         

            if predIndex == 11:
                return self.precpred(self._ctx, 15)
         

    def eID_sempred(self, localctx, predIndex):
            if predIndex == 12:
                return self.precpred(self._ctx, 2)
         

            if predIndex == 13:
                return self.precpred(self._ctx, 1)
         







"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements the simulator
"""

import intrepyd as ip
import intrepyd.api
import intrepyd.trace

class Simulator(object):
    """
    A Simulator
    """
    def __init__(self, ctx):
        self.ctx = ctx
        self.simulator = ip.api.mk_simulator(self.ctx)

    def add_watch(self, net):
        """
        Tells the simulator about a net whose value will be reported in the simulated trace
        """
        ip.api.simulator_add_watch(self.ctx, self.simulator, net)

    def simulate(self, trace, depth):
        """
        Executes a simulation using the values in trace, up to the specified depth
        """
        ip.api.simulator_simulate(self.simulator, trace.rawtrace, depth)



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017
"""
import intrepyd as ip
from intrepyd.engine import EngineResult
import unittest

class TestBr(unittest.TestCase):

    def test_br_01(self):
        context = ip.Context()
        realtype = context.mk_real_type()
        limit = context.mk_number("50", realtype)
        one = context.mk_number("1", realtype)
        counter1 = context.mk_latch("counter1", realtype)
        plusone = context.mk_add(counter1, one)
        context.set_latch_init_next(counter1, one, plusone)
        counter2 = context.mk_latch("counter2", realtype)
        plusone = context.mk_add(counter2, one)
        context.set_latch_init_next(counter2, one, plusone)
        mul = context.mk_mul(counter1, counter2)
        context.mk_output(mul)
        target = context.mk_leq(limit, mul)
        br = context.mk_backward_reach()
        br.add_target(target)
        br.add_watch(counter1)
        br.add_watch(counter2)
        result = br.reach_targets()
        self.assertEquals(EngineResult.REACHABLE, result)
        trace = br.get_last_trace()
        traceDict = trace.get_as_net_dictionary()
        counter1lastValue = traceDict[counter1][-1]
        counter2lastValue = traceDict[counter2][-1]
        self.assertEquals(counter1lastValue, "8.0")
        self.assertEquals(counter2lastValue, "8.0")

    def test_br_02(self):
        context = ip.Context()
        bt = context.mk_boolean_type()
        x = context.mk_input("x", bt)
        y = context.mk_input("y", bt)
        target1 = context.mk_and(x, y)
        target2 = context.mk_and(context.mk_not(x), y)
        br = context.mk_backward_reach()
        br.add_target(target1)
        br.add_target(target2)
        result = br.reach_targets()
        self.assertEquals(EngineResult.REACHABLE, result)
        self.assertEquals(1, len(list(br.get_last_reached_targets())))
        br.remove_last_reached_targets()
        result = br.reach_targets()
        self.assertEquals(EngineResult.REACHABLE, result)
        self.assertEquals(1, len(list(br.get_last_reached_targets())))
        br.remove_last_reached_targets()

    def test_br_03(self):
        context = ip.Context()
        rt = context.mk_real_type()
        x = context.mk_input("x", rt)
        y = context.mk_input("y", rt)
        mul = context.mk_mul(x, y)
        mul = context.mk_mul(mul, y)
        n70 = context.mk_number('70', rt)
        lt = context.mk_lt(n70, mul)
        br = context.mk_backward_reach()
        br.add_target(lt)
        result = br.reach_targets()
        self.assertEquals(EngineResult.REACHABLE, result)

if __name__ == '__main__':
    unittest.main()



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements infrastructure to store expressions
"""

from intrepyd.lustre2py.visitable import Visitable

class Expression(Visitable):
    """
    Abstract class for expressions
    """
    def __init__(self, kind):
        self._kind = kind

    @property
    def kind(self):
        """
        Getter
        """
        return self._kind

    INITCURR = 'INITCURR'
    LITERAL = 'LITERAL'
    ZEROARY = 'ZEROARY'
    UNARY = 'UNARY'
    BINARY = 'BINARY'
    CALL = 'CALL'
    TUPLE = 'TUPLE'
    ITE = 'ITE'

    INT = 'int'
    REAL = 'real'
    BOOL = 'bool'

class CallExpression(Expression):
    """
    Class for node calls
    """
    def __init__(self, cid, params):
        Expression.__init__(self, Expression.CALL)
        self._cid = cid
        self._params = params

    @property
    def cid(self):
        """
        Getter
        """
        return self._cid

    @property
    def params(self):
        """
        Getter
        """
        return self._params


class InitCurrExpression(Expression):
    """
    Class for init curr
    """
    def __init__(self, init, curr):
        Expression.__init__(self, Expression.INITCURR)
        self._init = init
        self._curr = curr

    @property
    def init(self):
        """
        Getter
        """
        return self._init

    @property
    def curr(self):
        """
        Getter
        """
        return self._curr

class LiteralExpression(Expression):
    """
    Class for numeric or boolean literals
    """
    def __init__(self, value, datatype):
        Expression.__init__(self, Expression.LITERAL)
        self._value = value
        self._datatype = datatype

    @property
    def value(self):
        """
        Getter
        """
        return self._value

    @property
    def datatype(self):
        """
        Getter
        """
        return self._datatype

class ZeroaryExpression(Expression):
    """
    Class for zero-arity functions
    """
    def __init__(self, name):
        Expression.__init__(self, Expression.ZEROARY)
        self._name = name

    @property
    def name(self):
        """
        Getter
        """
        return self._name

class UnaryExpression(Expression):
    """
    Class for unary functions
    """
    def __init__(self, operator, operand):
        Expression.__init__(self, Expression.UNARY)
        self._operator = operator
        self._operand = operand

    @property
    def operator(self):
        """
        Getter
        """
        return self._operator

    @property
    def operand(self):
        """
        Getter
        """
        return self._operand

class BinaryExpression(Expression):
    """
    Class for binary functions
    """
    def __init__(self, operator, left, right):
        Expression.__init__(self, Expression.BINARY)
        self._operator = operator
        self._left = left
        self._right = right

    @property
    def operator(self):
        """
        Getter
        """
        return self._operator


    @property
    def left(self):
        """
        Getter
        """
        return self._left

    @property
    def right(self):
        """
        Getter
        """
        return self._right

class ITEExpression(Expression):
    """
    Class for if then elses
    """
    def __init__(self, if_, then_, else_):
        Expression.__init__(self, Expression.ITE)
        self._if_ = if_
        self._then_ = then_
        self._else_ = else_

    @property
    def if_(self):
        """
        Getter
        """
        return self._if_

    @property
    def then_(self):
        """
        Getter
        """
        return self._then_

    @property
    def else_(self):
        """
        Getter
        """
        return self._else_


class TupleExpression(Expression):
    """
    Class for tuples
    """
    def __init__(self, expressions):
        Expression.__init__(self, Expression.TUPLE)
        self._expressions = expressions

    @property
    def expressions(self):
        """
        Getter
        """
        return self._expressions



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements plotting services
"""

import intrepyd as ip
import intrepyd.trace
from matplotlib import pyplot as plt

def create_subplot(allPlots, n, x, y, legend):
    y = [int(elem) for elem in y]
    miny = min(y)
    maxy = max(y)
    subplot = plt.subplot(allPlots, 1, n)
    subplot.axes.set_xticks(x)
    subplot.axes.set_yticks(range(miny, maxy + 1))
    subplot.set_ylim(miny - 1, maxy + 1)
    plt.step(x, y, label=legend)
    plt.legend(bbox_to_anchor=(1.05, 1), loc=2, borderaxespad=0.)
    # plt.plot()

def plot_trace_dictionary(trace):
    """
    Draws one step plot per each signal in the trace.

    Args:
        trace (Dictionary): the dictionary of the trace
    """
    if trace == None:
        raise Exception('Null trace given as input')

    if len(trace.keys()) == 0:
        return

    # Computes the step numerals [0, ...]
    steps = [step for step in range(len(trace[trace.keys()[0]]) + 1)]

    # Preprocess the counterexample duplicate the first value.
    # This is a visual trick to make plots look nicer (in particular
    # to have a better displaying of the first value that would
    # otherwise be hidden by the y-axis. Also turn 'true' in 1,
    # 'false' in 0, '?' in 0
    for key in trace.keys():
        if len(trace[key]) == 0:
            raise Exception('Unexpected trace with no values')
        for value in range(len(trace[key])):
            v = trace[key][value]
            v = ip.trace.Trace.get_numeric_value(v)
            if v == '?':
                trace[key][value] = 0
            else:
                trace[key][value] = v
        trace[key].insert(0, trace[key][0])

    # Create the plots, one in each subplot
    allPlots = len(trace.keys())
    n = 1
    for key in trace.keys():
       create_subplot(allPlots, n, steps, trace[key], key) 
       n += 1

    plt.tight_layout()
    plt.show()



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017
"""
import unittest
import intrepyd as ip
import intrepyd.atg
import intrepyd.circuit
import collections
import pandas as pd

class CircAnd(ip.circuit.Circuit):
    def __init__(self, ctx, name):
        ip.circuit.Circuit.__init__(self, ctx, name)

    def _mk_inputs(self):
        self.inputs['A'] = self.context.mk_input('A', self.context.mk_boolean_type())
        self.inputs['B'] = self.context.mk_input('B', self.context.mk_boolean_type())

    def _mk_naked_circuit_impl(self, inputs):
        n1 = self.inputs['A']
        n2 = self.inputs['B']
        outputs = collections.OrderedDict()
        out = self.context.mk_and(n1, n2)
        outputs['O'] = out
        self.nets['A'] = n1
        self.nets['B'] = n2
        self.nets['O'] = out
        return outputs

class TestAtg(unittest.TestCase):

    def test_atg_01(self):
        ctx = ip.Context()
        decisions = { 'O' : ['A', 'B'] }
        tables, decision2unreachable = ip.atg.compute_mcdc(ctx, CircAnd, decisions, maxDepth=10)
        decision2dataframe = ip.atg.get_tables_as_dataframe(tables)
        self.assertEqual(3, len(decision2dataframe['O']))
        self.assertEqual(0, len(decision2unreachable['O']))


"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017

This module implements the verification engines
"""

from enum import Enum
import intrepyd as ip

class EngineResult(Enum):
    """
    Results from engine's reachability calls
    """
    REACHABLE = ip.api.INT_ENGINE_RESULT_REACHABLE
    UNREACHABLE = ip.api.INT_ENGINE_RESULT_UNREACHABLE
    UNKNOWN = ip.api.INT_ENGINE_RESULT_UNKNOWN


class Engine(object):
    """
    Abstract intreface for an Engine
    """

    def __init__(self, ctx):
        self.ctx = ctx
        self.last_result = EngineResult.UNKNOWN

    def add_target(self, net):
        """
        Adds a target to be solved
        """
        raise NotImplementedError('Should have implemented this')

    def reach_targets(self):
        """
        Tries to reach any of the previously added targets
        """
        self.last_result = self._api_result_to_engine_result(self._reach_targets_impl())
        return self.last_result

    def get_last_reached_targets(self):
        """
        Returns the list of targets that has been reached last
        """
        if self.last_result != EngineResult.REACHABLE:
            return
        last_reached_targets_no = self._get_last_reached_targets_number_impl()
        return (self._get_last_reached_target_impl(i) for i in range(last_reached_targets_no))

    def get_last_trace(self):
        """
        Returns the last produced trace
        """
        raise NotImplementedError('Should have implemented this')

    def remove_last_reached_targets(self):
        """
        Removes all the reached targets from the last call from the engine
        """
        return self._remove_last_reached_targets_impl()

    def add_watch(self, net):
        """
        Track the value of net in future traces
        """
        raise NotImplementedError('Should have implemented this')

    def can_prove(self):
        """
        Tells if the engine can prove targets to be unreachable
        """
        raise NotImplementedError('Should have implemented this')

    def can_optimize(self):
        """
        Tells if the engine can prove targets to be unreachable
        """
        raise NotImplementedError('Should have implemented this')

    def _api_result_to_engine_result(self, result):
        if result == ip.api.INT_ENGINE_RESULT_REACHABLE:
            return EngineResult.REACHABLE
        elif result == ip.api.INT_ENGINE_RESULT_UNREACHABLE:
            return EngineResult.UNREACHABLE
        return EngineResult.UNKNOWN

    def _reach_targets_impl(self):
        """
        Engine-dependent reachability search for the added targets
        """
        raise NotImplementedError('Should have implemented this')

    def _get_last_reached_targets_number_impl(self):
        """
        Engine-dependent retrieval of last reached targets number
        """
        raise NotImplementedError('Should have implemented this')

    def _get_last_reached_target_impl(self, target):
        """
        Engine-dependent retrieval of last reached targets
        """
        raise NotImplementedError('Should have implemented this')

    def _remove_last_reached_targets_impl(self):
        """
        Engine-dependent removal of last reached targets
        """
        raise NotImplementedError('Should have implemented this')

class Bmc(Engine):
    """
    A Bounded Model Checking engine
    """

    def __init__(self, ctx):
        Engine.__init__(self, ctx)
        self.bmc = ip.api.mk_engine_bmc(self.ctx)

    def add_target(self, net):
        ip.api.bmc_add_target(self.ctx, self.bmc, net)

    def get_last_trace(self):
        if self.last_result != EngineResult.REACHABLE:
            raise Exception('Cannot get a trace as last result was not REACHABLE')
        target = next(self.get_last_reached_targets())
        cex = ip.api.bmc_get_trace(self.ctx, self.bmc, target)
        return ip.trace.Trace(self.ctx, cex)

    def add_watch(self, net):
        ip.api.bmc_add_watch(self.ctx, self.bmc, net)

    def set_current_depth(self, depth):
        """
        Sets the current depth to use for BMC
        """
        ip.api.set_bmc_current_depth(self.bmc, depth)

    def can_prove(self):
        return False

    def can_optimize(self):
        return False

    def _reach_targets_impl(self):
        return ip.api.bmc_reach_targets(self.bmc)

    def _get_last_reached_targets_number_impl(self):
        return ip.api.bmc_last_reached_targets_number(self.bmc)

    def _get_last_reached_target_impl(self, target):
        return ip.api.bmc_last_reached_target(self.bmc, target)

    def _remove_last_reached_targets_impl(self):
        return ip.api.bmc_remove_last_reached_targets(self.bmc)

class OptimizingBmc(Bmc):
    """
    An Optimizing Bounded Model Checking engine
    """

    def __init__(self, ctx):
        Bmc.__init__(self, ctx)
        ip.api.set_bmc_optimize(self.bmc)

    def can_optimize(self):
        return True


class BackwardReach(Engine):
    """
    A Backward Reachability Model Checking algorithm
    """

    def __init__(self, ctx):
        Engine.__init__(self, ctx)
        self.breach = ip.api.mk_engine_br(self.ctx)

    def add_target(self, net):
        ip.api.br_add_target(self.ctx, self.breach, net)

    def get_last_trace(self):
        if self.last_result != EngineResult.REACHABLE:
            raise Exception('Cannot get a trace as last result was not REACHABLE')
        target = next(self.get_last_reached_targets())
        rawtrace = ip.api.br_get_trace(self.ctx, self.breach, target)
        return ip.trace.Trace(self.ctx, rawtrace)

    def add_watch(self, net):
        ip.api.br_add_watch(self.ctx, self.breach, net)

    def can_prove(self):
        return True

    def can_optimize(self):
        return False

    def _reach_targets_impl(self):
        return ip.api.br_reach_targets(self.breach)

    def _get_last_reached_targets_number_impl(self):
        return ip.api.br_last_reached_targets_number(self.breach)

    def _get_last_reached_target_impl(self, target):
        return ip.api.br_last_reached_target(self.breach, target)

    def _remove_last_reached_targets_impl(self):
        return ip.api.br_remove_last_reached_targets(self.breach)



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 05/04/2017

This module implements a printer of an AST into Intrepyd syntax
"""

from intrepyd.lustre2py.visitor import Visitor
from intrepyd.lustre2py.instruction import Equation, Property
from intrepyd.lustre2py.node import Node
from intrepyd.lustre2py.expression import Expression

CONTEXT = 'ctx'
FIRSTTICK = '__first_tick'
INTTYPE = '__it'
REALTYPE = '__rt'
BOOLTYPE = '__bt'
TAB = ' ' * 4
LATCH2PRE = 'LATCH2PRE'
LATCHEQUIV = 'LATCHEQUIV'
_DEFAULT_BOOL_INIT = CONTEXT + '.mk_false()'
_DEFAULT_ARIT_INIT = CONTEXT + '.mk_number("0", '

LUSTREOP2INTREPYDUNARYOP = {
    '-' : 'mk_minus',
    'not' : 'mk_not'
}

LUSTREOP2INTREPYDBINARYOP = {
    '+' : 'mk_add',
    '-' : 'mk_sub',
    '*' : 'mk_mul',
    '/' : 'mk_div',
    '=' : 'mk_eq',
    '<>' : 'mk_neq',
    '<' : 'mk_lt',
    '>' : 'mk_gt',
    '<=' : 'mk_leq',
    '>=' : 'mk_geq',
    '=>' : 'mk_implies',
    'div' : 'mk_div',
    'mod' : 'mk_mod',
    'or' : 'mk_or',
    'and' : 'mk_and',
    'xor' : 'mk_xor',
}

LUSTREDT2INTREPYDDT = {
    'bool' : BOOLTYPE,
    'int' : INTTYPE,
    'real' : REALTYPE
}

def _infer_datatype(expression, var2datatype, node2proto):
    if expression.kind == Expression.LITERAL:
        if expression.datatype == Expression.INT:
            return INTTYPE
        elif expression.datatype == Expression.REAL:
            return REALTYPE
        elif expression.datatype == Expression.BOOL:
            return BOOLTYPE
        raise TypeError('Unhandled type' + expression.datatype)
    elif expression.kind == Expression.ZEROARY:
        if not expression.name in var2datatype:
            raise RuntimeError('Unknown type for ' + expression.name)
        name = var2datatype[expression.name].name
        assert name in LUSTREDT2INTREPYDDT
        return LUSTREDT2INTREPYDDT[name]
    elif expression.kind == Expression.UNARY:
        if expression.operator == 'not':
            return BOOLTYPE
        elif expression.operator == 'real':
            return REALTYPE
        elif expression.operator == 'floor':
            return INTTYPE
        else:
            # Recursive call
            return _infer_datatype(expression.operand, var2datatype, node2proto)
    elif expression.kind == Expression.BINARY:
        if expression.operator in set(['and', 'or', 'xor', '=>', '<', '>', '=', '<>', '<=', '>=']):
            return BOOLTYPE
        else:
            # Recursive call
            return _infer_datatype(expression.left, var2datatype, node2proto)
    elif expression.kind == Expression.ITE:
        # Recursive call
        return _infer_datatype(expression.then_, var2datatype, node2proto)
    elif expression.kind == Expression.INITCURR:
        # Recursive call
        return _infer_datatype(expression.init, var2datatype, node2proto)
    elif expression.kind == Expression.CALL:
        assert expression.cid in node2proto
        # outputs = [1], first output = [0]
        ttype = node2proto[expression.cid][1][0]
        assert ttype in LUSTREDT2INTREPYDDT
        return LUSTREDT2INTREPYDDT[ttype]
    elif expression.kind == Expression.TUPLE:
        assert len(expression.expressions) == 1
        # Recursive call
        return _infer_datatype(expression.expressions[0], var2datatype, node2proto)
    raise TypeError('Unhandled kind' + expression.kind)

def _contains(expression, var):
    if expression.kind == Expression.LITERAL:
        return False
    elif expression.kind == Expression.ZEROARY:
        return expression.name == var
    elif expression.kind == Expression.UNARY:
        return _contains(expression.operand, var)
    elif expression.kind == Expression.BINARY:
        return _contains(expression.left, var) or _contains(expression.right, var)
    elif expression.kind == Expression.ITE:
        return _contains(expression.if_, var) or\
               _contains(expression.then_, var) or\
               _contains(expression.else_, var)
    elif expression.kind == Expression.INITCURR:
        return _contains(expression.init, var) or\
               _contains(expression.next, var)
    elif expression.kind == Expression.CALL:
        for param in expression.params:
            if param == var:
                return True
        return False
    elif expression.kind == Expression.TUPLE:
        for expr in expression.expressions:
            if _contains(expr, var):
                return True
        return False
    raise TypeError('Unhandled kind' + expression.kind)


class Ast2Intrepyd(Visitor):
    """
    Visitor for printing AST on stdout
    """
    def __init__(self, node2proto, realtype):
        self._result = ''
        self._main_node = None
        self._properties = []
        self._id = 0
        self._nid = 0
        self._var2datatype = {}
        self._latch_decls = []
        self._node2proto = node2proto
        self._pre_decls = set()
        self._realtype = realtype

    def _get_new_name(self):
        self._id += 1
        return '__l' + str(self._id)

    def _get_unique_namespace(self, name):
        self._nid += 1
        return '"%s.%d"' % (name, self._nid)

    def _visit_node(self, node):
        assert isinstance(node, Node)
        self._pre_decls = set()
        self._var2datatype = {}
        self._latch_decls = []
        if node.main:
            self._main_node = node
        self._result += 'def ' + node.name + '(' + CONTEXT + ', ' + FIRSTTICK
        for inp in node.input_decls:
            self._result += ', ' + inp.accept(self)
        self._result += '):\n'
        self._result += TAB + BOOLTYPE + ' = ' + CONTEXT + '.mk_boolean_type()\n'
        self._result += TAB + INTTYPE + ' = ' + CONTEXT + '.mk_int32_type()\n'
        self._result += TAB + REALTYPE + ' = ' + CONTEXT + '.mk_' + self._realtype + '_type()\n'
        for loc in node.local_decls:
            loc.accept(self)
        for out in node.output_decls:
            out.accept(self)
        for equat in node.equations:
            equat.accept(self)
        for latch_decl in self._latch_decls:
            new_name, init, operand = latch_decl
            self._result += TAB + 'if ' + operand + ' in ' + LATCH2PRE + ':\n'
            pair = '(' + LATCH2PRE + '[' + operand + '], ' + new_name + ')'
            self._result += TAB + TAB + LATCHEQUIV + '.append(' + pair + ')\n'
            self._result += TAB + LATCH2PRE + '[' + operand + '] = ' + new_name + '\n'
            self._result += TAB + CONTEXT +\
                            '.set_latch_init_next(' + new_name + ', ' +\
                                                  init + ', ' + operand + ')\n'
        for prop in node.properties:
            self._properties.append(prop.accept(self))
        sep = ''
        for out in node.output_decls:
            for var in out.variables:
                self._result += TAB + CONTEXT + '.net2name[' + var + '] = "' + var + '"\n'
        self._result += TAB + 'return '
        for out in node.output_decls:
            self._result += sep + out.accept(self)
            sep = ', '
        self._result += '\n\n'
        return self._result, self._properties

    def _visit_var_decl_list(self, lst):
        result = ''
        sep = ''
        for grp in lst.var_decl_groups:
            result += sep + grp.accept(self)
            sep = ', '
        return result

    def _visit_var_decl_group(self, grp):
        result = ''
        sep = ''
        for var in grp.variables:
            result += sep + var
            sep = ', '
            self._var2datatype[var] = grp.datatype
        return result

    def _visit_equation(self, equat):
        assert isinstance(equat, Equation)
        expr = equat.expression.accept(self)
        result = TAB
        sep = ''
        for lhs in equat.lhs:
            result += sep + lhs
            sep = ', '
        result += ' = ' + expr + '\n'
        self._result += result

    def _visit_init_curr_expression(self, expr):
        new_name = self._get_new_name()
        init = expr.init.accept(self)
        curr = expr.curr.accept(self)
        self._result += TAB + new_name + ' = ' +\
                        CONTEXT + '.mk_ite(' +\
                                  FIRSTTICK + ', ' +\
                                  init + ', ' + curr + ')\n'
        return new_name

    def _visit_literal_expression(self, expr):
        if expr.value == 'true':
            return CONTEXT + '.mk_true()'
        if expr.value == 'false':
            return CONTEXT + '.mk_false()'
        datatype = LUSTREDT2INTREPYDDT[expr.datatype]
        return CONTEXT + '.mk_number("' + expr.value + '", ' + datatype + ')'

    def _visit_zeroary_expression(self, expr):
        return expr.name

    def _visit_unary_expression(self, expr):
        operand = expr.operand.accept(self)
        if expr.operator == 'pre':
            new_name = 'pre_' + operand
            # return latch created in this function somewhere else
            if new_name in self._pre_decls:
                return new_name
            self._pre_decls.add(new_name)
            datatype = _infer_datatype(expr.operand, self._var2datatype, self._node2proto)
            self._result += TAB + new_name + ' = ' + CONTEXT +\
                            '.mk_latch("' + new_name + '", ' + datatype + ')\n'
            init = None
            if datatype == BOOLTYPE:
                init = _DEFAULT_BOOL_INIT
            else:
                init = _DEFAULT_ARIT_INIT + datatype + ')'
            self._latch_decls.append((new_name, init, operand))
            return new_name
        new_name = self._get_new_name()
        self._result += TAB + new_name + ' = '
        assert expr.operator in LUSTREOP2INTREPYDUNARYOP
        self._result += CONTEXT + '.' +\
                        LUSTREOP2INTREPYDUNARYOP[expr.operator] + '(' + operand + ')\n'
        return new_name

    def _visit_binary_expression(self, expr):
        new_name = self._get_new_name()
        left = expr.left.accept(self)
        right = expr.right.accept(self)
        self._result += TAB + new_name + ' = '
        assert expr.operator in LUSTREOP2INTREPYDBINARYOP
        self._result += CONTEXT + '.' + LUSTREOP2INTREPYDBINARYOP[expr.operator] +\
                        '(' + left + ', ' + right + ')\n'
        return new_name

    def _visit_ite_expression(self, expr):
        new_name = self._get_new_name()
        if_ = expr.if_.accept(self)
        then_ = expr.then_.accept(self)
        else_ = expr.else_.accept(self)
        self._result += TAB + new_name + ' = ' + CONTEXT +\
                        '.mk_ite(' + if_ + ', ' + then_ + ', ' + else_ + ')\n'
        return new_name

    def _visit_call_expression(self, expr):
        new_name = self._get_new_name()
        result = TAB + CONTEXT +\
                 '.push_namespace(' + self._get_unique_namespace(expr.cid) + ')\n'
        result += TAB + new_name + ' = ' + expr.cid
        result += '(' + CONTEXT + ', ' + FIRSTTICK
        for param in expr.params:
            param_name = param.accept(self)
            result += ', ' + param_name
        result += ')\n'
        self._result += result
        self._result += TAB + CONTEXT + '.pop_namespace()\n'
        return new_name

    def _visit_primitive(self, obj):
        return obj.name

    def _visit_tuple_expression(self, ttuple):
        put_par = len(ttuple.expressions) > 1
        result = ''
        if put_par:
            result += '('
        sep = ''
        for expr in ttuple.expressions:
            result += sep + expr.accept(self)
            sep = ', '
        if put_par:
            result += ')'
        return result

    def _visit_enum(self, obj):
        raise NotImplementedError

    def _visit_struct(self, obj):
        raise NotImplementedError

    def _visit_array(self, obj):
        raise NotImplementedError

    def _visit_subrange(self, obj):
        raise NotImplementedError

    def _visit_property(self, obj):
        assert isinstance(obj, Property)
        result = obj.expression.accept(self)
        self._properties.append(result)



"""
Copyright (C) 2017 Roberto Bruttomesso <roberto.bruttomesso@gmail.com>

This file is distributed under the terms of the 3-clause BSD License.
A copy of the license can be found in the root directory or at
https://opensource.org/licenses/BSD-3-Clause.

Author: Roberto Bruttomesso <roberto.bruttomesso@gmail.com>
  Date: 27/03/2017
"""
import intrepyd as ip
from intrepyd.engine import EngineResult
import intrepyd.pseudoboolean
import unittest

class TestExactlyOne(unittest.TestCase):

    def create_and_run_nets(self, n, context):
        nets = []
        bt = context.mk_boolean_type()
        for i in range(n):
            nets.append(context.mk_input("n" + str(i), bt))
        eo = ip.pseudoboolean.mk_exactly_one(context.ctx, nets, "eo")
        bmc = context.mk_bmc()
        bmc.add_target(eo)
        result = bmc.reach_targets()
        self.assertEqual(EngineResult.REACHABLE, result)
        reachedNets = bmc.get_last_reached_targets()
        self.assertEqual(1, len(list(reachedNets)))

    def test_exactly_one_01(self):
        context = ip.Context()
        self.create_and_run_nets(1, context)

    def test_exactly_one_02(self):
        context = ip.Context()
        self.create_and_run_nets(2, context)

    def test_exactly_one_03(self):
        context = ip.Context()
        self.create_and_run_nets(3, context)

    def test_exactly_one_04(self):
        context = ip.Context()
        self.create_and_run_nets(4, context)

    def test_exactly_one_05(self):
        context = ip.Context()
        self.create_and_run_nets(10, context)

    def test_exactly_one_06(self):
        context = ip.Context()
        self.create_and_run_nets(11, context)

    def test_exactly_one_07(self):
        context = ip.Context()
        self.create_and_run_nets(12, context)

    def test_exactly_one_08(self):
        context = ip.Context()
        self.create_and_run_nets(43, context)

    def test_exactly_one_09(self):
        context = ip.Context()
        nets = []
        bt = context.mk_boolean_type()
        i1 = context.mk_input("i1", bt)
        i2 = context.mk_input("i2", bt)
        i3 = context.mk_input("i3", bt)
        nets.append(context.mk_or(i1, i2))
        nets.append(context.mk_or(i1, i3))
        nets.append(context.mk_or(i2, i3))
        eo = ip.pseudoboolean.mk_exactly_one(context.ctx, nets, "eo")
        bmc = context.mk_bmc()
        bmc.add_target(eo)
        result = bmc.reach_targets()
        self.assertEqual(EngineResult.UNREACHABLE, result)

if __name__ == "__main__":
    unittest.main()
